"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains algorithms for data analysis.

Code written by: Luc IJspeert
"""
import numpy as np

from star_shine.core import time_series as tms, periodogram as pdg, fitting as fit
from star_shine.core import frequency_sets as frs, utility as ut
from star_shine.config.helpers import get_config


# load configuration
config = get_config()


def extract_single(time, flux, f0=-1, fn=-1, select='a'):
    """Extract a single sinusoid from a time series.

    The extracted frequency is based on the highest amplitude or signal-to-noise in the periodogram.
    The highest peak is oversampled by a factor 100 to get a precise measurement.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f0: float, optional
        Lowest allowed frequency for extraction.
        If left -1, default is f0 = 1/(100*T)
    fn: float, optional
        Highest allowed frequency for extraction.
        If left -1, default is fn = 1/(2*np.min(np.diff(time))) = Nyquist frequency
    select: str, optional
        Select the next frequency based on amplitude 'a' or signal-to-noise 'sn'

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f_final: float
            Frequency of the extracted sinusoid
        a_final: float
            Amplitude of the extracted sinusoid
        ph_final: float
            Phase of the extracted sinusoid

    See Also
    --------
    scargle, scargle_phase_single
    """
    df = 0.1 / np.ptp(time)  # default frequency sampling is about 1/10 of frequency resolution

    # full LS periodogram
    freqs, ampls = pdg.scargle_parallel(time, flux, f0=f0, fn=fn, df=df)

    # selection step based on flux to noise (refine step keeps using ampl)
    if select == 'sn':
        noise_spectrum = pdg.scargle_noise_spectrum_redux(freqs, ampls, window_width=1.0)
        ampls = ampls / noise_spectrum

    # select highest amplitude
    i_f_max = np.argmax(ampls)

    # refine frequency by increasing the frequency resolution x100
    f_left = max(freqs[i_f_max] - df, df / 10)  # may not get too low
    f_right = freqs[i_f_max] + df
    f_refine, a_refine = pdg.scargle(time, flux, f0=f_left, fn=f_right, df=df/100)

    # select refined highest amplitude
    i_f_max = np.argmax(a_refine)
    f_final = f_refine[i_f_max]
    a_final = a_refine[i_f_max]

    # finally, compute the phase (and make sure it stays within + and - pi)
    _, ph_final = pdg.scargle_ampl_phase_single(time, flux, f_final)

    return f_final, a_final, ph_final


def extract_local(time, flux, f0, fn):
    """Extract a single sinusoid from a time series at a predefined frequency interval.

    The extracted frequency is based on the highest amplitude in the periodogram.
    The highest peak is oversampled by a factor 100 to get a precise measurement.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f0: float
        Lowest allowed frequency for extraction.
    fn: float
        Highest allowed frequency for extraction.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f_final: float
            Frequency of the extracted sinusoid
        a_final: float
            Amplitude of the extracted sinusoid
        ph_final: float
            Phase of the extracted sinusoid

    See Also
    --------
    scargle, scargle_phase_single
    """
    df = 0.1 / np.ptp(time)  # default frequency sampling is about 1/10 of frequency resolution

    # partial LS periodogram
    freqs, ampls = pdg.scargle(time, flux, f0=f0, fn=fn, df=df)

    # cut off the ends of the frequency range if they are rising
    i_f_min_edges = ut.uphill_local_max(freqs, -ampls, freqs[np.array([0, -1])])
    freqs = freqs[i_f_min_edges[0]:i_f_min_edges[1] + 1]
    ampls = ampls[i_f_min_edges[0]:i_f_min_edges[1] + 1]

    # select highest amplitude
    i_f_max = np.argmax(ampls)

    # refine frequency by increasing the frequency resolution x100
    f_left = max(freqs[i_f_max] - df, df / 10)  # may not get too low
    f_right = freqs[i_f_max] + df
    f_refine, a_refine = pdg.scargle(time, flux, f0=f_left, fn=f_right, df=df / 100)

    # select refined highest amplitude
    i_f_max = np.argmax(a_refine)
    f_final = f_refine[i_f_max]
    a_final = a_refine[i_f_max]

    # finally, compute the phase (and make sure it stays within + and - pi)
    _, ph_final = pdg.scargle_ampl_phase_single(time, flux, f_final)

    return f_final, a_final, ph_final


def extract_approx(time, flux, f_approx):
    """Extract a single sinusoid from a time series at an approximate location.

    Follows the periodogram upwards to the nearest peak. The periodogram is oversampled for a more precise result.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    f_approx: float
        Approximate location of the frequency of maximum amplitude.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        f_final: float
            Frequency of the extracted sinusoid
        a_final: float
            Amplitude of the extracted sinusoid
        ph_final: float
            Phase of the extracted sinusoid
    """
    df = 0.1 / np.ptp(time)  # default frequency sampling is about 1/10 of frequency resolution

    # LS periodogram around the approximate location (2.5 times freq res)
    f0 = max(f_approx - 25 * df, df / 10)
    fn = f_approx + 25 * df
    freqs, ampls = pdg.scargle(time, flux, f0=f0, fn=fn, df=df)

    # get the index of the frequency of the maximum amplitude
    i_f_max = ut.uphill_local_max(freqs, ampls, np.array([f_approx]))[0]

    # refine frequency by increasing the frequency resolution x100
    f_left = max(freqs[i_f_max] - df, df / 10)
    f_right = freqs[i_f_max] + df
    f_refine, a_refine = pdg.scargle(time, flux, f0=f_left, fn=f_right, df=df / 100)

    # select refined highest amplitude
    i_f_max = np.argmax(a_refine)
    f_final = f_refine[i_f_max]
    a_final = a_refine[i_f_max]

    # finally, compute the phase
    _, ph_final = pdg.scargle_ampl_phase_single(time, flux, f_final)

    return f_final, a_final, ph_final


def refine_subset(ts_model, close_f, logger=None):
    """Refine a subset of frequencies that are within the Rayleigh criterion of each other,
    taking into account (and not changing the frequencies of) harmonics if present.

    Intended as a sub-loop within another extraction routine (extract_sinusoids), can work standalone too.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    close_f: numpy.ndarray[Any, dtype[int]]
        Indices of the subset of frequencies to be refined
    logger: logging.Logger, optional
        Instance of the logging library.

    See Also
    --------
    extract_sinusoids
    """
    # get all the harmonics (regardless of base)
    i_harmonics = np.arange(len(ts_model.sinusoid.harmonics))[ts_model.sinusoid.harmonics]

    # determine initial bic
    bic_prev = ts_model.bic()

    # stop the loop when the BIC increases
    condition_1 = True
    while condition_1:
        # remember the current sinusoids
        f_c, a_c, ph_c = ts_model.sinusoid.get_sinusoid_parameters(exclude=False)

        # remove each frequency one at a time to then re-extract them
        for j in close_f:
            # exclude the sinusoid
            ts_model.exclude_sinusoids(j)
            # update the linear model for good measure
            ts_model.update_linear_model()

            # improve sinusoid j by re-extracting its parameters
            f_j = ts_model.sinusoid.f_n[j]
            if j in i_harmonics:
                # if f is a harmonic, don't shift the frequency
                a_j, ph_j = pdg.scargle_ampl_phase_single(ts_model.time, ts_model.residual(), f_j)
            else:
                f_j, a_j, ph_j = extract_approx(ts_model.time, ts_model.residual(), f_j)

            # update the model
            ts_model.update_sinusoids(f_j, a_j, ph_j, j)

        # as a last model-refining step, redetermine the constant and slope
        ts_model.update_linear_model()

        # calculate BIC before moving to the next iteration
        bic = ts_model.bic()
        d_bic = bic_prev - bic

        # stop the loop when the BIC increases
        condition_1 = np.round(d_bic, 2) > 0

        # check acceptance condition before moving to the next iteration
        if condition_1:
            # accept the new frequency
            bic_prev = bic
        else:
            # update the sinusoids back to their original values
            ts_model.update_sinusoids(f_c, a_c, ph_c, close_f)
            ts_model.update_linear_model()

    if logger is not None:
        logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic_prev:1.2f} - N_refined= {len(close_f)}",
                     extra={'update': True})

    return None

def replace_subset(ts_model, close_f, final_remove=True, logger=None):
    """Attempt the replacement of frequencies within the Rayleigh criterion of each other by a single one,
    taking into account (and not changing the frequencies of) harmonics if present.

    Intended as a sub-loop within another extraction routine (extract_sinusoids), can work standalone too.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    close_f: numpy.ndarray[Any, dtype[int]]
        Indices of the subset of frequencies to be (subdivided and) replaced.
    final_remove: bool, optional
        Remove the excluded frequencies at the end.
    logger: logging.Logger, optional
        Instance of the logging library.

    See Also
    --------
    extract_sinusoids
    """
    # standard frequency resolution (not the user defined one)
    freq_res = 1 / ts_model.t_tot
    # get all the harmonics (regardless of base)
    i_harmonics = np.arange(len(ts_model.sinusoid.harmonics))[ts_model.sinusoid.harmonics]

    # make all combinations of consecutive frequencies in close_f (longer sets first)
    close_f_sets = ut.consecutive_subsets(close_f)

    # determine initial quantities
    n_sin_tot_init = len(ts_model.sinusoid.f_n)
    n_excluded_init = np.sum(~ts_model.sinusoid.include)
    bic_prev = ts_model.bic()

    # loop over all subsets:
    for set_i in close_f_sets:
        # if set_i contains removed sinusoids, skip (order of sets matters)
        if np.any(~ts_model.sinusoid.include[set_i]):
            continue

        # exclude the next set of sinusoids
        ts_model.exclude_sinusoids(set_i)
        # update the linear model for good measure
        ts_model.update_linear_model()

        # check for harmonics
        harm_i = [h for h in set_i if h in i_harmonics]

        # remove all frequencies in the set and re-extract one
        f_c = ts_model.sinusoid.f_n  # current frequencies
        if len(harm_i) > 0:
            # if f is a harmonic, don't shift the frequency
            f_i = f_c[harm_i]  # can be more than one harmonic
            a_i, ph_i = pdg.scargle_ampl_phase(ts_model.time, ts_model.residual(), f_i)
        else:
            f0 = min(f_c[set_i]) - freq_res
            fn = max(f_c[set_i]) + freq_res
            f_i, a_i, ph_i = extract_local(ts_model.time, ts_model.residual(), f0=f0, fn=fn)

        # add sinusoid to the model
        ts_model.add_sinusoids(f_i, a_i, ph_i)
        # as a last model-refining step, redetermine the constant and slope
        ts_model.update_linear_model()

        # calculate BIC before moving to the next iteration
        bic = ts_model.bic()
        d_bic = bic_prev - bic

        # acceptance condition for replacement
        condition_1 = np.round(d_bic, 2) > 0

        # check acceptance condition before moving to the next iteration
        if condition_1:
            # accept the changes
            bic_prev = bic
        else:
            # remove the added sinusoid(s)
            ts_model.remove_sinusoids(np.arange(len(f_c), len(f_c) + len(np.atleast_1d(f_i))))

            # include the excluded sinusoids
            ts_model.include_sinusoids(set_i)
            ts_model.update_linear_model()

    # determine number of excluded
    n_excluded = np.sum(~ts_model.sinusoid.include[:n_sin_tot_init])
    n_replaced = n_excluded - n_excluded_init

    if final_remove:
        ts_model.remove_excluded()

    if logger is not None:
        logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic_prev:1.2f} - N_replaced= {n_replaced}",
                     extra={'update': True})

    return None


def extract_sinusoids(ts_model, bic_thr=2, snr_thr=0, stop_crit='bic', select='hybrid', n_extract=0,
                      fit_each_step=False, replace_each_step=True, logger=None):
    """Extract all the frequencies from a periodic flux.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    bic_thr: float, optional
        The minimum decrease in BIC by fitting a sinusoid for the signal to be considered significant.
    snr_thr: float, optional
        Threshold for signal-to-noise ratio for a signal to be considered significant.
    stop_crit: str, optional
        Use the BIC as stopping criterion or the SNR, choose from 'bic', or 'snr'
    select: str, optional
        Select the next frequency based on amplitude ('a'),
        signal-to-noise ('sn'), or hybrid ('hybrid') (first a then sn).
    n_extract: int, optional
        Maximum number of frequencies to extract. The stop criterion is still leading. Zero means as many as possible.
    fit_each_step: bool
        If set to True, a non-linear least-squares fit of all extracted sinusoids in groups is performed at each
        iteration. While this increases the quality of the extracted signals, it drastically slows down the code.
        Generally gives a better quality of the extraction than only doing this all the way at the end.
    replace_each_step: bool
        If set to True, close frequecies are attempted to be replaced by a single sinusoid at each iteration.
        May increase the quality of the extraction more than only doing this all the way at the end.
    logger: logging.Logger, optional
        Instance of the logging library.

    Notes
    -----
    Spits out frequencies and amplitudes in the same units as the input,
    and phases that are measured with respect to the first time point.
    Also determines the flux average, so this does not have to be subtracted
    before input into this function.
    Note: does not perform a non-linear least-squares fit at the end,
    which is highly recommended! (In fact, no fitting is done at all).

    The function optionally takes a pre-existing frequency list to append
    additional frequencies to. Set these to np.array([]) to start from scratch.

    i_chunks is a 2D array with start and end indices of each (half) sector.
    This is used to model a piecewise-linear trend in the data.
    If you have no sectors like the TESS mission does, set
    i_chunks = np.array([[0, len(time)]])

    Exclusively uses the Lomb-Scargle periodogram (and an iterative parameter
    improvement scheme) to extract the frequencies.
    Uses a delta BIC > bic_thr stopping criterion.

    [Author's note] Although it is my belief that doing a non-linear
    multi-sinusoid fit at each iteration of the prewhitening is the
    ideal approach, it is also a very (very!) time-consuming one and this
    algorithm aims to be fast while approaching the optimal solution.

    [Another author's note] I added an option to do the non-linear multi-
    sinusoid fit at each iteration.
    """
    if n_extract == 0:
        n_extract = 10**6  # 'a lot'

    # initial number of sinusoids
    n_sin_init = ts_model.sinusoid.n_sin

    # set up selection process
    if select == 'hybrid':
        switch = True  # when we would normally end, we switch strategy
        select = 'a'  # start with amplitude extraction
    else:
        switch = False

    # determine the initial bic
    bic_prev = ts_model.bic()  # initialise current BIC to the mean (and slope) subtracted flux

    # log a message
    if logger is not None:
        logger.extra(f"N_f= {n_sin_init}, BIC= {bic_prev:1.2f} - Iterative prewhitening")

    # stop the loop when the BIC decreases by less than 2 (or increases)
    condition_1 = True
    condition_2 = True
    while (condition_1 | switch) & condition_2:
        # switch selection method when extraction would normally stop
        if switch and not condition_1:
            select = 'sn'
            switch = False
            logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic_prev:1.2f} - Switch selection from a to sn.")

        # remember the current sinusoids
        f_c, a_c, ph_c = ts_model.sinusoid.get_sinusoid_parameters(exclude=False)
        _, h_base, h_mult = ts_model.sinusoid.get_harmonic_parameters(exclude=False)

        # attempt to extract the next frequency
        f_i, a_i, ph_i = extract_single(ts_model.time, ts_model.residual(), f0=ts_model.pd_f0, fn=ts_model.pd_fn,
                                        select=select)
        ts_model.add_sinusoids(f_i, a_i, ph_i)

        # update the linear pars
        ts_model.update_linear_model()

        # improve sinusoids with some strategy
        if fit_each_step:
            # fit all sinusoids for best improvement
            fit.fit_multi_sinusoid_grouped(ts_model, logger=logger)
        else:
            # select only close frequencies for iteration
            close_f = frs.f_within_rayleigh(ts_model.sinusoid.n_sin - 1, ts_model.sinusoid.f_n, ts_model.f_resolution)

            if len(close_f) > 1:
                # iterate over (re-extract) close frequencies (around f_i) a number of times to improve them
                refine_subset(ts_model, close_f, logger=logger)

        # possibly replace close frequencies
        if replace_each_step:
            close_f = frs.f_within_rayleigh(ts_model.sinusoid.n_sin - 1, ts_model.sinusoid.f_n, ts_model.f_resolution)
            if len(close_f) > 1:
                replace_subset(ts_model, close_f, logger=logger)

        # calculate BIC
        bic = ts_model.bic()
        d_bic = bic_prev - bic

        # acceptance condition
        if stop_crit == 'snr':
            # calculate SNR in a 1 c/d window around the extracted frequency
            noise = pdg.scargle_noise_at_freq(np.array([f_i]), ts_model.time, ts_model.residual(), window_width=1.0)
            snr = a_i / noise
            # stop the loop if snr threshold not met
            condition_1 = snr > snr_thr
        else:
            # stop the loop when the BIC decreases by less than bic_thr (or increases)
            condition_1 = np.round(d_bic, 2) > bic_thr

        # check acceptance condition before moving to the next iteration
        if condition_1:
            # accept the new frequency
            bic_prev = bic
        else:
            ts_model.set_sinusoids(f_c, a_c, ph_c, h_base_new=h_base, h_mult_new=h_mult)
            ts_model.update_linear_model()

        # stop the loop if n_sin reaches limit
        condition_2 = ts_model.sinusoid.n_sin - n_sin_init < n_extract

        if logger is not None and condition_1:
            logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic:1.2f} - Extracted: f= {f_i:1.6f}, a= {a_i:1.6f}",
                         extra={'update': True})

    if logger is not None:
        n_sin = ts_model.sinusoid.n_sin
        logger.extra(f"N_f= {n_sin}, BIC= {bic_prev:1.2f} - N_extracted= {n_sin - n_sin_init}.")

    return None


def couple_harmonics(ts_model, f_base, logger=None):
    """Finds matching harmonics in the sinusoids and couples their frequencies.

    Surrounding frequencies are also removed (within the frequency resolution).

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    f_base: float
        Base frequency to couple the harmonics to.
    logger: logging.Logger, optional
        Instance of the logging library.
    """
    # find the harmonic candidates using the period
    harmonics, h_mult = frs.find_harmonics_tolerance(ts_model.sinusoid.f_n, f_base, f_tol=ts_model.f_resolution / 2)

    if len(harmonics) == 0:
        if logger is not None:
            logger.warning("No harmonic frequencies found")

        return None

    # the index of f_base is len(f_n), because the base harmonic is the first frequency to be added at the end
    i_base = len(ts_model.sinusoid.f_n)

    # if the base frequency is not yet present, add it
    if 1 not in h_mult:
        a_1, ph_1 = pdg.scargle_ampl_phase_single(ts_model.time, ts_model.residual(), f_base)
        ts_model.add_sinusoids(f_base, a_1, ph_1, h_base_new=i_base, h_mult_new=1)

    # determine initial quantities
    n_harm_init = len(harmonics)

    # go through the harmonics by harmonic number and re-extract them (n==1 must come first, if present)
    for n in np.unique(h_mult):
        # get the indices to exclude, disregarding included state
        n_sin_tot = len(ts_model.sinusoid.f_n)
        remove = np.arange(n_sin_tot)[harmonics][h_mult == n]

        # exclude the neighbouring harmonic candidates and update linear model
        ts_model.exclude_sinusoids(remove)
        ts_model.update_linear_model()

        # extract the harmonic amplitude and phase
        f_i = n * f_base
        a_i, ph_i = pdg.scargle_ampl_phase_single(ts_model.time, ts_model.residual(), f_i)

        # add harmonic candidate and redetermine the constant and slope
        ts_model.add_sinusoids(f_i, a_i, ph_i, h_base_new=i_base, h_mult_new=n)

    # lastly re-determine slope and const and remove the excluded frequencies
    ts_model.remove_excluded()
    ts_model.update_linear_model()

    if logger is not None:
        bic = ts_model.bic()
        logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic:1.2f} - N_coupled= {ts_model.sinusoid.n_harm}, "
                     f"N_harmonics_init= {n_harm_init}", extra={'update': True})

    return None


def extract_harmonics(ts_model, bic_thr=2, logger=None):
    """Tries to extract more harmonics from the flux

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    bic_thr: float
        The minimum decrease in BIC by fitting a sinusoid for the signal to be considered significant.
    logger: logging.Logger, optional
        Instance of the logging library.

    See Also
    --------
    fix_harmonic_frequency

    Notes
    -----
    Looks for missing harmonics and checks whether adding them decreases the BIC sufficiently (by more than 2).
    Assumes the harmonics are already fixed multiples of f_base as can be achieved with fix_harmonic_frequency.
    """
    # make lists of not-present possible harmonics paired with their base frequency
    i_base_all = []
    f_base_all = []
    h_candidates_n = []
    for i_base in np.unique(ts_model.sinusoid.h_base[ts_model.sinusoid.harmonics]):
        # the range of harmonic multipliers below the Nyquist frequency
        harmonics_i = np.arange(1, ts_model.pd_fn / ts_model.sinusoid.f_n[i_base], dtype=int)

        # h_mult minus one is the position for existing harmonics
        harmonics_i = np.delete(harmonics_i, ts_model.sinusoid.h_mult[ts_model.sinusoid.h_base == i_base] - 1)
        h_candidates_n.extend(harmonics_i)

        # add to the bae frequency lists
        i_base_all.extend([i_base for _ in range(len(harmonics_i))])
        f_base_all.extend([ts_model.sinusoid.f_n[i_base] for _ in range(len(harmonics_i))])

    # determine initial quantities
    n_sin_init = ts_model.sinusoid.n_sin
    bic_prev = ts_model.bic()  # initialise current BIC to the mean (and slope) subtracted flux

    if logger is not None:
        logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic_prev:1.2f} - Extract harmonics")

    # loop over candidates and try to extract (BIC decreases by 2 or more)
    for i in range(len(i_base_all)):
        f_i = h_candidates_n[i] * f_base_all[i]
        a_i, ph_i = pdg.scargle_ampl_phase_single(ts_model.time, ts_model.residual(), f_i)

        # add harmonic candidate and redetermine the constant and slope
        ts_model.add_sinusoids(f_i, a_i, ph_i, h_base_new=i_base_all[i], h_mult_new=h_candidates_n[i])
        ts_model.update_linear_model()

        # determine new BIC and whether it improved
        bic = ts_model.bic()
        d_bic = bic_prev - bic

        # stop the loop when the BIC decreases by less than bic_thr (or increases)
        condition_1 = np.round(d_bic, 2) > bic_thr

        # check acceptance condition before moving to the next iteration
        if condition_1:
            # accept the new frequency
            bic_prev = bic
        else:
            # h_c is rejected, revert to previous model
            ts_model.remove_sinusoids(len(ts_model.sinusoid.f_n) - 1)  # remove last added
            ts_model.update_linear_model()

        if logger is not None and condition_1:
            logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic:1.2f} - Extracted: "
                         f"f_base= {f_base_all[i]:1.2f}, h= {h_candidates_n[i]}, f= {f_i:1.6f}, a= {a_i:1.6f}",
                         extra={'update': True})

    if logger is not None:
        n_sin = ts_model.sinusoid.n_sin
        logger.extra(f"N_f= {n_sin}, BIC= {bic_prev:1.2f} - N_h_extracted= {n_sin - n_sin_init}")

    return None


def replace_close_sinusoid_pair(ts_model):
    """Replace pairs of sinusoids that have gotten too close.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    """
    f_n = ts_model.sinusoid.f_n
    harmonics = ts_model.sinusoid.harmonics

    # get the pairs of close frequencies
    i_close = frs.f_close_pairs(f_n, df=ts_model.pd_df)

    # replace each pair, taking harmonics into account
    for pair in i_close:
        # exclude the pair
        ts_model.exclude_sinusoids(pair)

        # check for harmonic
        if np.any(harmonics[pair]):
            # if f is a harmonic, don't shift the frequency
            f_i = f_n[pair][harmonics[pair]]  # select the harmonic
            a_i, ph_i = pdg.scargle_ampl_phase(ts_model.time, ts_model.residual(), f_i)
        else:
            f0 = f_n[pair[0]] - ts_model.pd_df  # pairs are ordered
            fn = f_n[pair[1]] + ts_model.pd_df  # pairs are ordered
            f_i, a_i, ph_i = extract_local(ts_model.time, ts_model.residual(), f0=f0, fn=fn)

        # add sinusoid to the model
        ts_model.add_sinusoids(f_i, a_i, ph_i)
        # as a last model-refining step, redetermine the constant and slope
        ts_model.update_linear_model()

    # remove the excluded sinusoids
    ts_model.remove_excluded()

    return None


def remove_sinusoids_single(ts_model, logger=None):
    """Attempt the removal of individual sinusoids.

    Checks whether the BIC can be improved by removing a sinusoid. Harmonics are taken into account.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    logger: logging.Logger, optional
        Instance of the logging library.
    """
    # determine initial quantities
    n_sin_init = len(ts_model.sinusoid.f_n)
    bic_prev = ts_model.bic()

    # while frequencies are added to the exclude list, continue loop
    n_sin = np.sum(ts_model.sinusoid.include)
    n_prev = n_sin + 1
    while n_sin < n_prev:
        n_prev = n_sin
        for i in range(n_sin_init):
            # continue if sinusoid is already excluded, or when it is a base harmonic
            if not ts_model.sinusoid.include[i] or ts_model.sinusoid.h_mult[i] == 1:
                continue

            # exclude sinusoid and redetermine the constant and slope
            ts_model.exclude_sinusoids(i)
            ts_model.update_linear_model()

            # determine new BIC and whether it improved
            bic = ts_model.bic()
            d_bic = bic_prev - bic

            # only remove sinusoid if it increases the BIC
            condition_1 = np.round(d_bic, 2) > 0

            # check acceptance condition before moving to the next iteration
            if condition_1:
                # accept the removal
                bic_prev = bic
                n_sin = np.sum(ts_model.sinusoid.include)
            else:
                # removal is rejected, revert to previous model
                ts_model.include_sinusoids(i)

    # lastly re-determine slope and const and remove the excluded frequencies
    ts_model.remove_excluded()
    ts_model.update_linear_model()

    if logger is not None:
        n_sin = ts_model.sinusoid.n_sin
        logger.extra(f"N_f= {n_sin}, BIC= {bic_prev:1.2f} - N_removed= {n_sin_init - n_sin}.", extra={'update': True})

    return None


def replace_sinusoid_groups(ts_model, logger=None):
    """Attempt the replacement of groups of frequencies by a single one.

    Checks whether the BIC can be improved by replacing a group of frequencies by only one.
    Harmonics are never removed.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    logger: logging.Logger, optional
        Instance of the logging library.
    """
    # make an array of sets of frequencies to be investigated for replacement
    close_f_groups = frs.chains_within_rayleigh(ts_model.sinusoid.f_n, ts_model.f_resolution)

    # determine initial quantities
    n_sin_tot_init = len(ts_model.sinusoid.f_n)
    n_excluded_init = np.sum(~ts_model.sinusoid.include)

    # while frequencies are added to the exclude list, continue loop
    n_sin = np.sum(ts_model.sinusoid.include)
    n_prev = n_sin + 1
    while n_sin < n_prev:
        n_prev = n_sin
        for i, close_f in enumerate(close_f_groups):
            # continue if full sinusoid set is already excluded
            if np.all(~ts_model.sinusoid.include[close_f]):
                continue

            # use the replace_subset function to handle the details
            replace_subset(ts_model, close_f, final_remove=False, logger=None)

            # update number of sinusoids after replacement
            n_sin = np.sum(ts_model.sinusoid.include)

    # determine number of excluded and added
    n_excluded = np.sum(~ts_model.sinusoid.include[:n_sin_tot_init])
    n_replaced = n_excluded - n_excluded_init
    n_new = len(ts_model.sinusoid.f_n) - n_sin_tot_init

    # lastly re-determine slope and const and remove the excluded frequencies
    ts_model.remove_excluded()
    ts_model.update_linear_model()

    if logger is not None:
        bic = ts_model.bic()
        logger.extra(f"N_f= {ts_model.sinusoid.n_sin}, BIC= {bic:1.2f} - N_replaced= {n_replaced}, N_kept= {n_new}.",
                     extra={'update': True})

    return None


def reduce_sinusoids(ts_model, logger=None):
    """Attempt to reduce the number of frequencies taking into account any harmonics if present.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    logger: logging.Logger, optional
        Instance of the logging library.

    Notes
    -----
    Checks whether the BIC can be improved by removing a frequency. Special attention is given to frequencies
    that are within the Rayleigh criterion of each other. It is attempted to replace these by a single frequency.
    """
    # first check if any frequency can be left out (after the fit, this may be possible)
    remove_sinusoids_single(ts_model, logger=logger)

    # Now go on to trying to replace sets of frequencies that are close together
    replace_sinusoid_groups(ts_model, logger=logger)

    return None


def select_sinusoids(ts_model, logger=None):
    """Selects the credible frequencies from the given set

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.
    logger: logging.Logger, optional
        Instance of the logging library.

    Notes
    -----
    Harmonic frequencies that are said to be passing the criteria are in fact passing the criteria for
    individual frequencies, not those for a set of harmonics (which would be a looser constraint).
    """
    n_sin = ts_model.sinusoid.n_sin

    # update the sinusoid errors in the model
    ts_model.update_sinusoid_uncertainties()
    ts_model.update_sinusoid_uncertainties_harmonic()

    # find the insignificant frequencies
    ts_model.update_sinusoid_passing_sigma()

    # apply the signal-to-noise threshold
    ts_model.update_sinusoid_passing_snr(window_width=config.window_width)

    # candidate harmonic frequencies passing criteria
    ts_model.update_sinusoid_passing_harmonic()

    if logger is not None:
        passed_all = (ts_model.sinusoid.passing_sigma & ts_model.sinusoid.passing_snr)
        n_pass_all = np.sum(passed_all)
        n_harm = np.sum(ts_model.sinusoid.passing_harmonic)
        n_harm_pass_all = np.sum(passed_all & ts_model.sinusoid.passing_harmonic)
        logger.extra(f"Sinusoids passing criteria: {n_pass_all} of {n_sin}. "
                     f"Harmonics passing criteria: {n_harm_pass_all} of {n_harm}.")

    return None


def refine_harmonic_base_frequency(f_base, ts_model):
    """Refine the base frequency for a harmonic sinusoid model.

    Parameters
    ----------
    f_base: float
        Base frequency of the harmonic series to refine.
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.

    Returns
    -------
    float
        Base frequency of the harmonic series.

    Notes
    -----
    Uses the sum of distances between the harmonics and their
    theoretical positions to refine the orbital period.

    Period precision is 0.00001, but accuracy is a bit lower.

    Same refine algorithm as used in find_orbital_period
    """
    freq_res = ts_model.f_resolution
    f_nyquist = ts_model.pd_fn
    f_n = ts_model.sinusoid.f_n

    # refine by using a dense sampling and the harmonic distances
    f_refine = np.arange(0.99 * f_base, 1.01 * f_base, 0.00001 * f_base)
    n_harm_r, completeness_r, distance_r = frs.harmonic_series_length(f_refine, f_n, freq_res, f_nyquist)
    h_measure = n_harm_r * completeness_r  # compute h_measure for constraining a domain
    mask_peak = (h_measure > np.max(h_measure) / 1.5)  # constrain the domain of the search
    i_min_dist = np.argmin(distance_r[mask_peak])
    f_base = f_refine[mask_peak][i_min_dist]

    return f_base


def find_harmonic_base_frequency(ts_model):
    """Find the most likely base frequency for a harmonic sinusoid model.

    Parameters
    ----------
    ts_model: tms.TimeSeriesModel
        Instance of TimeSeriesModel containing the time series and model parameters.

    Returns
    -------
    float
        Base frequency of the harmonic series.

    Notes
    -----
    Uses a combination of phase dispersion minimisation and Lomb-Scargle periodogram (see Saha & Vivas 2017),
    and some refining steps to get the best frequency.

    Also tests various multiples of the period. Precision is 0.00001 (one part in one-hundred-thousand).
    (accuracy might be slightly lower)
    """
    freq_res = ts_model.f_resolution
    f_nyquist = ts_model.pd_fn
    f_n = ts_model.sinusoid.f_n

    # first to get a global minimum do combined PDM and LS, at select frequencies
    periods, phase_disp = pdg.phase_dispersion_minimisation(ts_model.time, ts_model.flux, f_n, local=False)
    freqs = 1 / periods
    ampls, _ = pdg.scargle_ampl_phase(ts_model.time, ts_model.flux, freqs)
    psi_measure = ampls / phase_disp

    # also check the number of harmonics at each period and include into best f
    n_harm, completeness, distance = frs.harmonic_series_length(freqs, f_n, freq_res, f_nyquist)
    psi_h_measure = psi_measure * n_harm * completeness

    # select the best period, refine it and check double P
    f_base = freqs[np.argmax(psi_h_measure)]

    # refine by using a dense sampling and the harmonic distances
    f_refine = np.arange(0.99 * f_base, 1.01 * f_base, 0.00001 * f_base)
    n_harm_r, completeness_r, distance_r = frs.harmonic_series_length(f_refine, f_n, freq_res, f_nyquist)
    h_measure = n_harm_r * completeness_r  # compute h_measure for constraining a domain
    mask_peak = (h_measure > np.max(h_measure) / 1.5)  # constrain the domain of the search
    i_min_dist = np.argmin(distance_r[mask_peak])
    f_base = f_refine[mask_peak][i_min_dist]

    # reduce the search space by taking limits in the distance metric
    f_left = f_refine[mask_peak][:i_min_dist]
    f_right = f_refine[mask_peak][i_min_dist:]
    d_left = distance_r[mask_peak][:i_min_dist]
    d_right = distance_r[mask_peak][i_min_dist:]
    d_max = np.max(distance_r)
    if np.any(d_left > d_max / 2):
        f_l_bound = f_left[d_left > d_max / 2][-1]
    else:
        f_l_bound = f_refine[mask_peak][0]
    if np.any(d_right > d_max / 2):
        f_r_bound = f_right[d_right > d_max / 2][0]
    else:
        f_r_bound = f_refine[mask_peak][-1]
    bound_interval = f_r_bound - f_l_bound

    # decide on the multiple of the period
    harmonics, harmonic_n = frs.find_harmonics_from_pattern(f_n, f_base, f_tol=freq_res / 2)
    completeness_p = (len(harmonics) / (f_nyquist // f_base))
    completeness_p_l = (len(harmonics[harmonic_n <= 15]) / (f_nyquist // f_base))

    # check these (commonly missed) fractions
    n_multiply = np.array([1/2, 2, 3, 4, 5])
    f_fracs = f_base / n_multiply
    n_harm_r_m, completeness_r_m, distance_r_m = frs.harmonic_series_length(f_fracs, f_n, freq_res, f_nyquist)
    h_measure_m = n_harm_r_m * completeness_r_m  # compute h_measure for constraining a domain

    # if there are very high numbers, add double that fraction for testing
    test_frac = h_measure_m / h_measure[mask_peak][i_min_dist]
    if np.any(test_frac[2:] > 3):
        n_multiply = np.append(n_multiply, [2 * n_multiply[2:][test_frac[2:] > 3]])
        f_fracs = f_base / n_multiply
        n_harm_r_m, completeness_r_m, distance_r_m = frs.harmonic_series_length(f_fracs, f_n, freq_res, f_nyquist)
        h_measure_m = n_harm_r_m * completeness_r_m  # compute h_measure for constraining a domain

    # compute diagnostic fractions that need to meet some threshold
    test_frac = h_measure_m / h_measure[mask_peak][i_min_dist]
    compl_frac = completeness_r_m / completeness_p

    # doubling the period may be done if the harmonic filling factor below f_16 is very high
    f_cut = np.max(f_n[harmonics][harmonic_n <= 15])
    f_n_c = f_n[f_n <= f_cut]
    n_harm_r_2, completeness_r_2, distance_r_2 = frs.harmonic_series_length(f_fracs, f_n_c, freq_res, f_nyquist)
    compl_frac_2 = completeness_r_2[1] / completeness_p_l

    # empirically determined thresholds for the various measures
    minimal_frac = 1.1
    minimal_compl_frac = 0.85
    minimal_frac_low = 0.95
    minimal_compl_frac_low = 0.95

    # test conditions
    test_condition = (test_frac > minimal_frac)
    compl_condition = (compl_frac > minimal_compl_frac)
    test_condition_2 = (test_frac[1] > minimal_frac_low)
    compl_condition_2 = (compl_frac_2 > minimal_compl_frac_low)
    if np.any(test_condition & compl_condition) | (test_condition_2 & compl_condition_2):
        if np.any(test_condition & compl_condition):
            i_best = np.argmax(test_frac[compl_condition])
            f_base = f_fracs[compl_condition][i_best]
        else:
            f_base = f_base / 2

        # make new bounds for refining
        f_left_b = f_base - (bound_interval / 2)
        f_right_b = f_base + (bound_interval / 2)

        # refine by using a dense sampling and the harmonic distances
        f_refine_2 = np.arange(f_left_b, f_right_b, 0.00001 * f_base)
        n_harm_r2, completeness_r2, distance_r2 = frs.harmonic_series_length(f_refine_2, f_n, freq_res, f_nyquist)
        h_measure_2 = n_harm_r2 * completeness_r2  # compute h_measure for constraining a domain
        mask_peak = (h_measure_2 > np.max(h_measure_2) / 1.5)  # constrain the domain of the search
        i_min_dist = np.argmin(distance_r2[mask_peak])
        f_base = f_refine_2[mask_peak][i_min_dist]

    return f_base



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains functions that evaluate certain configuration attributes that depend on data properties.

Code written by: Luc IJspeert
"""
import numpy as np
import numba as nb

from star_shine.config.helpers import get_config


# load configuration
config = get_config()


def signal_to_noise_threshold(time):
    """Determine the signal-to-noise threshold for accepting frequencies based on the number of points

    Based on Baran & Koen 2021, eq 6. (https://ui.adsabs.harvard.edu/abs/2021AcA....71..113B/abstract)

    snr_thr = 1.201 * np.sqrt(1.05 * np.log(len(time)) + 7.184)
    Plus 0.25 in case of gaps > 27 days

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.

    Returns
    -------
    float
        Signal-to-noise threshold for this data set.
    """
    # if user defined (unequal to -1), return the configured number
    if config.snr_thr != -1:
        return config.snr_thr

    # equation 6 from Baran & Koen 2021
    snr_thr = 1.201 * np.sqrt(1.05 * np.log(len(time)) + 7.184)

    # increase threshold by 0.25 if gaps longer than 27 days
    if np.any((time[1:] - time[:-1]) > 27):
        snr_thr += 0.25

    # round to two decimals
    snr_thr = np.round(snr_thr, 2)

    return snr_thr


def frequency_resolution(time):
    """Calculate the frequency resolution of a time series

    Equation: factor / T, where T is the total time base of observations.
    Recommended factor: 1.5. Common choices are 1, 1.5 (conservative), 2.5 (very conservative).

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.

    Returns
    -------
    float
        Frequency resolution of the time series
    """
    f_res = config.resolution_factor / np.ptp(time)

    return f_res


@nb.njit(cache=True)
def nyquist_sum_koen_2006(n, time, delta_t_min):
    """Calculate the Nyquist sum based on Koen (2006).

    Parameters
    ----------
    n: int
        The number of terms in the series.
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    delta_t_min: float
        Minimum time interval between observations.

    Returns
    -------
    int
        Result of the sum of squares calculation.

    Examples
    --------
    # Iterate n until this sum returns zero to get to the true Nyquist frequency.
    >>> delta_t_min = np.min(time[1:] - time[:-1])
    >>> precision = 1e-10
    >>> ss_nu = 1e-9
    >>> n = 0
    >>> while (ss_nu > precision):
    >>>     n += 1
    >>>     ss_nu = nyquist_sum_koen_2006(n, time, delta_t_min)
    >>>
    >>> f_nyquist = n / (2 * delta_t_min)
    # Note that this may go on for a while.

    Notes
    -----
    The principle exploited by Koen (2006) is that the real Nyquist frequency of a time series with uneven sampling,
    where the sampling is also not a simple multiple of some small value, is a multiple of the simple approximation
    of the Nyquist frequency:
    f = 1 / (2 * delta_t_min)
    Since this function is computationally intensive, it is recommended to simply pick a multiple of the
    simple approximation in case a higher value is desired.
    """
    factor = n * np.pi / delta_t_min

    # evaluate equation 5 from Koen 2006 at nu = 2pi*n/delta_t_min
    ss = 0
    for i in range(0, len(time) - 1):
        for j in range(i + 1, len(time)):
            ss += np.sin(factor * (time[j] - time[i]))**2

    return ss


def nyquist_frequency(time):
    """Determines the maximum frequency for extraction and periodograms.

    The Nyquist frequency is calculated using the configured built-in function.
    The rigorous method implements equation 5 from Koen 2006.
    https://ui.adsabs.harvard.edu/abs/2006MNRAS.371.1390K/abstract

    Currently limited to 20 times classical Nyquist for performance considerations.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.

    Returns
    -------
    float
        Nyquist frequency of the time series
    """
    # calculate the Nyquist frequency
    f_nyquist = config.nyquist_factor / (2 * np.min(time[1:] - time[:-1]))

    return f_nyquist



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the settings dialog for the graphical user interface.

Code written by: Luc IJspeert
"""
from PySide6.QtWidgets import QDialog, QHBoxLayout, QVBoxLayout, QFormLayout, QLineEdit, QPushButton, QLabel, QCheckBox
from PySide6.QtWidgets import QMessageBox, QFrame

from star_shine.api.main import update_config, save_config
from star_shine.config.helpers import get_config


# load configuration
config = get_config()


class SettingsDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.config = config
        self.setWindowTitle("Settings")

        layout = QHBoxLayout()
        left_layout = QVBoxLayout()

        # create the data and file settings form
        left_layout.addWidget(QLabel("Data and File Settings"))
        data_file_form_layout = self._create_data_file_settings()
        left_layout.addLayout(data_file_form_layout)

        # Create a horizontal line as a divider
        h_line = QFrame()
        h_line.setFrameShape(QFrame.HLine)
        h_line.setFrameShadow(QFrame.Sunken)
        left_layout.addWidget(h_line)

        # create the tabulated file settings form
        left_layout.addWidget(QLabel("Tabulated File Settings"))
        data_file_form_layout = self._create_tabulated_settings()
        left_layout.addLayout(data_file_form_layout)

        # Create a horizontal line as a divider
        h_line = QFrame()
        h_line.setFrameShape(QFrame.HLine)
        h_line.setFrameShadow(QFrame.Sunken)
        left_layout.addWidget(h_line)

        # create the fits file settings form
        left_layout.addWidget(QLabel("Fits File Settings"))
        data_file_form_layout = self._create_fits_settings()
        left_layout.addLayout(data_file_form_layout)

        layout.addLayout(left_layout)

        # Create a vertical line as a divider
        v_line = QFrame()
        v_line.setFrameShape(QFrame.VLine)
        v_line.setFrameShadow(QFrame.Sunken)
        layout.addWidget(v_line)

        # create the GUI settings on the right side
        right_layout = QVBoxLayout()

        # create the extraction settings form
        right_layout.addWidget(QLabel("Extraction Settings"))
        extraction_form_layout = self._create_extraction_settings()
        right_layout.addLayout(extraction_form_layout)

        # Create a horizontal line as a divider
        h_line = QFrame()
        h_line.setFrameShape(QFrame.HLine)
        h_line.setFrameShadow(QFrame.Sunken)
        right_layout.addWidget(h_line)

        # create the GUI settings form
        right_layout.addWidget(QLabel("GUI Settings"))
        gui_form_layout = self._create_gui_settings()
        right_layout.addLayout(gui_form_layout)

        # the save/cancel buttons
        button_box = QHBoxLayout()
        apply_button = QPushButton("Apply")
        apply_button.clicked.connect(self.apply_settings)
        button_box.addWidget(apply_button)

        save_button = QPushButton("Save")
        save_button.clicked.connect(self.save_settings)
        button_box.addWidget(save_button)

        cancel_button = QPushButton("Cancel")
        cancel_button.clicked.connect(self.reject)
        button_box.addWidget(cancel_button)

        right_layout.addLayout(button_box)

        layout.addLayout(right_layout)

        self.setLayout(layout)

    def _create_extraction_settings(self):
        """Create the settings form for the extraction settings."""
        form_layout = QFormLayout()

        # Create input fields for each setting
        self.select_next_field = QLineEdit(self.config.select_next)
        form_layout.addRow(QLabel("Select Next Sinusoid:"), self.select_next_field)

        self.optimise_step_field = QCheckBox()
        self.optimise_step_field.setChecked(self.config.optimise_step)
        form_layout.addRow(QLabel("Optimise Every Step:"), self.optimise_step_field)

        self.replace_step_field = QCheckBox()
        self.replace_step_field.setChecked(self.config.replace_step)
        form_layout.addRow(QLabel("Replace Every Step:"), self.replace_step_field)

        self.bic_thr_field = QLineEdit(str(self.config.bic_thr))
        form_layout.addRow(QLabel("BIC Threshold:"), self.bic_thr_field)

        self.snr_thr_field = QLineEdit(str(self.config.snr_thr))
        form_layout.addRow(QLabel("SNR Threshold:"), self.snr_thr_field)

        self.nyquist_factor_field = QLineEdit(str(self.config.nyquist_factor))
        form_layout.addRow(QLabel("Nyquist Factor:"), self.nyquist_factor_field)

        self.resolution_factor_field = QLineEdit(str(self.config.resolution_factor))
        form_layout.addRow(QLabel("Resolution Factor:"), self.resolution_factor_field)

        self.window_width_field = QLineEdit(str(self.config.window_width))
        form_layout.addRow(QLabel("Window Width:"), self.window_width_field)

        return form_layout

    def _create_data_file_settings(self):
        """Create the settings form for the data and file settings."""
        form_layout = QFormLayout()

        self.overwrite_field = QCheckBox()
        self.overwrite_field.setChecked(self.config.overwrite)
        form_layout.addRow(QLabel("Overwrite files:"), self.overwrite_field)

        # data_dir is not used in the gui for simplicity

        self.save_dir_field = QLineEdit(self.config.save_dir)
        form_layout.addRow(QLabel("Save Directory:"), self.save_dir_field)

        self.save_ascii_field = QCheckBox()
        self.save_ascii_field.setChecked(self.config.save_ascii)
        form_layout.addRow(QLabel("Save ASCII files:"), self.save_ascii_field)

        return form_layout

    def _create_tabulated_settings(self):
        """Create the settings form for the tabulated file settings."""
        form_layout = QFormLayout()

        self.cn_time_field = QLineEdit(self.config.cn_time)
        form_layout.addRow(QLabel("Column Name For time:"), self.cn_time_field)

        self.cn_flux_field = QLineEdit(self.config.cn_flux)
        form_layout.addRow(QLabel("Column Name For flux:"), self.cn_flux_field)

        self.cn_flux_err_field = QLineEdit(self.config.cn_flux_err)
        form_layout.addRow(QLabel("Column Name For flux_err:"), self.cn_flux_err_field)

        return form_layout

    def _create_fits_settings(self):
        """Create the settings form for the FITS file settings."""
        form_layout = QFormLayout()

        self.cf_time_field = QLineEdit(self.config.cf_time)
        form_layout.addRow(QLabel("Column Name For time:"), self.cf_time_field)

        self.cf_flux_field = QLineEdit(self.config.cf_flux)
        form_layout.addRow(QLabel("Column Name For flux:"), self.cf_flux_field)

        self.cf_flux_err_field = QLineEdit(self.config.cf_flux_err)
        form_layout.addRow(QLabel("Column Name For flux_err:"), self.cf_flux_err_field)

        self.cf_quality_field = QLineEdit(self.config.cf_quality)
        form_layout.addRow(QLabel("Column Name For quality:"), self.cf_quality_field)

        self.apply_q_flags_field = QCheckBox()
        self.apply_q_flags_field.setChecked(self.config.apply_q_flags)
        form_layout.addRow(QLabel("Apply Quality Flags:"), self.apply_q_flags_field)

        self.halve_chunks_field = QCheckBox()
        self.halve_chunks_field.setChecked(self.config.halve_chunks)
        form_layout.addRow(QLabel("Halve Time Chunks:"), self.halve_chunks_field)

        return form_layout

    def _create_gui_settings(self):
        """Create the settings form for the GUI settings."""
        form_layout = QFormLayout()

        self.dark_mode_field = QCheckBox()
        self.dark_mode_field.setChecked(False)
        form_layout.addRow(QLabel("Dark Mode [WIP]:"), self.dark_mode_field)

        self.h_size_frac_field = QLineEdit(str(self.config.h_size_frac))
        form_layout.addRow(QLabel("Horizontal Size Fraction:"), self.h_size_frac_field)

        self.v_size_frac_field = QLineEdit(str(self.config.v_size_frac))
        form_layout.addRow(QLabel("Vertical Size Fraction:"), self.v_size_frac_field)

        return form_layout

    def apply_settings(self):
        """Apply the settings to the configuration"""
        try:
            setting_dict = {}

            # make a settings dictionary
            setting_dict['overwrite'] = self.overwrite_field.isChecked()
            setting_dict['save_dir'] = self.save_dir_field.text()
            setting_dict['save_ascii'] = self.save_ascii_field.isChecked()

            setting_dict['cn_time'] = self.cn_time_field.text()
            setting_dict['cn_flux'] = self.cn_flux_field.text()
            setting_dict['cn_flux_err'] = self.cn_flux_err_field.text()

            setting_dict['cf_time'] = self.cf_time_field.text()
            setting_dict['cf_flux'] = self.cf_flux_field.text()
            setting_dict['cf_flux_err'] = self.cf_flux_err_field.text()
            setting_dict['cf_quality'] = self.cf_quality_field.text()
            setting_dict['apply_q_flags'] = self.apply_q_flags_field.isChecked()
            setting_dict['halve_chunks'] = self.halve_chunks_field.isChecked()

            setting_dict['select_next'] = self.select_next_field.text()
            setting_dict['optimise_step'] = self.optimise_step_field.isChecked()
            setting_dict['replace_step'] = self.replace_step_field.isChecked()
            setting_dict['bic_thr'] = float(self.bic_thr_field.text())
            setting_dict['snr_thr'] = float(self.snr_thr_field.text())
            setting_dict['nyquist_factor'] = float(self.nyquist_factor_field.text())
            setting_dict['resolution_factor'] = float(self.resolution_factor_field.text())
            setting_dict['window_width'] = float(self.window_width_field.text())

            setting_dict['dark_mode'] = self.dark_mode_field.isChecked()
            setting_dict['h_size_frac'] = float(self.h_size_frac_field.text())
            setting_dict['v_size_frac'] = float(self.v_size_frac_field.text())

            # Update the configuration with new values
            update_config(settings=setting_dict)
        except ValueError:
            QMessageBox.warning(self, "Input Error", "Invalid input for settings.")

        # Close the dialog
        self.accept()

    def save_settings(self):
        """Save the settings form to disk."""
        self.apply_settings()

        try:
            # Save the configuration to a file
            save_config()
        except ValueError:
            QMessageBox.warning(self, "IO Error", "Error while saving config.")



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the analysis functions for the graphical user interface.

Code written by: Luc IJspeert
"""
from PySide6.QtCore import QThread, Signal


class PipelineThread(QThread):
    """A QThread subclass to perform analysis in the background."""

    def __init__(self, pipeline_instance):
        super().__init__()
        self.pipeline_instance = pipeline_instance
        self.func_name = ''
        self.args = ()
        self.kwargs = {}

    def start_function(self, func_name, *args, **kwargs):
        """Start the thread with a specific function and arguments."""
        self.func_name = func_name
        self.args = args
        self.kwargs = kwargs
        super().start()

    def run(self):
        """Run the function in a separate thread."""
        if self.func_name != '':
            # try:
            # start a thread with the function
            function_to_run = getattr(self.pipeline_instance, self.func_name)
            function_to_run(*self.args, **self.kwargs)

            # except Exception as e:
            #     self.pipeline_instance.logger.error(f"Error during analysis: {e}")

    def stop(self):
        """Stop the thread."""
        pass



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains utility functions for data processing, unit conversions.

Code written by: Luc IJspeert
"""
import datetime
import numpy as np
import numba as nb

from star_shine.config.helpers import get_config


# load configuration
config = get_config()


def datetime_formatted():
    """Return datetime string without microseconds.

    Returns
    -------
    str
        Date and time without microseconds
    """
    dt = datetime.datetime.now()
    dt_str = str(dt.date()) + ' ' + str(dt.hour) + ':' + str(dt.minute) + ':' + str(dt.second)

    return dt_str


@nb.njit(cache=True)
def decimal_figures(x, n_sf):
    """Determine the number of decimal figures to print given a target
    number of significant figures

    Parameters
    ----------
    x: float
        Value to determine the number of decimals for.
    n_sf: int
        Number of significant figures to compute.

    Returns
    -------
    int
        Number of decimal places to round to.
    """
    if x != 0:
        decimals = (n_sf - 1) - int(np.floor(np.log10(abs(x))))
    else:
        decimals = 1

    return decimals


@nb.njit(cache=True)
def float_to_str_numba(x, dec=2):
    """Convert float to string for Numba up to some decimal place
    
    Parameters
    ----------
    x: float
        Value to convert
    dec: int
        Number of decimals (be careful with large numbers here)
    
    Returns
    -------
    str
        String with the value x
    """
    x_round = np.round(x, dec)
    x_int = int(x_round)
    x_dec = int(np.abs(np.round(x_round - x_int, dec)) * 10**dec)
    s = str(x_int) + '.' + str(x_dec).zfill(dec)

    return s


def float_to_str_scientific(x, x_err, error=True, brackets=False):
    """Conversion of a number with an error margin to string in scientific notation.

    Uses two significant figures by default as no distinction is made between having a 1, 2 or higher number
    in the error value.

    Parameters
    ----------
    x: float
        Value to determine the number of decimals for.
    x_err: float
        Value to determine the number of decimals for.
    error: bool, optional
        Include the error value.
    brackets: bool, optional
        Place the error value in brackets.

    Returns
    -------
    str
        Formatted string conversion.
    """
    # determine the decimal places to round to
    rnd_x = max(decimal_figures(x_err, 2), decimal_figures(x, 2))

    # format the error value
    if brackets:
        err_str = f"(\u00B1{x_err:.{rnd_x}f})"
    else:
        err_str = f"\u00B1 {x_err:.{rnd_x}f}"

    # format the rest of the string
    if error:
        number_str = f"{x:.{rnd_x}f} {err_str}"
    else:
        number_str = f"{x:.{rnd_x}f}"

    return number_str


@nb.njit(cache=True)
def weighted_mean(x, w):
    """Weighted mean since Numba doesn't support numpy.average
    
    Parameters
    ----------
    x: numpy.ndarray[Any, dtype[float]]
        Values to calculate the mean over
    w: numpy.ndarray[Any, dtype[float]]
        Weights corresponding to each value
    
    Returns
    -------
    float
        Mean of x weighted by w
    """
    w_mean = np.sum(x * w) / np.sum(w)

    return w_mean


@nb.njit(cache=True)
def std_unb(x, n):
    """Unbiased standard deviation

    Parameters
    ----------
    x: numpy.ndarray[Any, dtype[float]]
        Values to calculate the std over
    n: int
        Number of degrees of freedom

    Returns
    -------
    float
        Unbiased standard deviation
    """
    residuals = x - np.mean(x)

    # tested to be faster in numba than np.sum(x**2)
    sum_r_2 = 0
    for i in range(len(residuals)):
        sum_r_2 += residuals[i]**2

    std = np.sqrt(sum_r_2 / n)  # unbiased standard deviation of the residuals

    return std


def consecutive_subsets(x):
    """Creates all consecutive subsets of the given list.

    Parameters
    ----------
    x: list[Any], numpy.ndarray[Any, dtype[Any]]
        A list of values.

    Returns
    -------
    list
        The list of consecutive subsets ordered by size and with length two or more.
    """
    n = len(x)

    # create the subsets from largest to smallest
    subsets = [x[p1:p1 + l] for l in range(n, 1, -1) for p1 in range(n - l + 1)]

    return subsets


def adjust_indices_removed(x, removed):
    """Adjusts the indices in `x` to account for removals.

    If y_reduced=np.delete(y, removed), and x are indices indicating items in y, this function gives
    the indices x_adj that indicate the same items in y_reduced. Assumes that none of the removed items
    are indicated by indices in x.

    Parameters
    ----------
    x: iterable[int], numpy.ndarray[Any, dtype[int]]
        A list of indices that slice another array, y.
    removed: iterable[int], numpy.ndarray[Any, dtype[int]]
        A second list of indices indicating removals from y.

    Returns
    -------
    list[int]
        Adjusted indices that can be used with the reduced y.
    """
    # adjust the indices
    x_adj = [i - sum(1 for r in removed if r < i) for i in x]

    return x_adj


@nb.njit(cache=True)
def find_local_max(y):
    """Find the indeces of the local maxima in `y`.

    The array `x` must be sorted in ascending order. This function identifies local maxima in the `y` array
    and returns the corresponding indeces.

    Parameters
    ----------
    y: numpy.ndarray[Any, dtype[float]]
        1D array of y-values corresponding to the x-values.

    Returns
    -------
    numpy.ndarray[Any, dtype[int]]
        Indeces of the `x` values corresponding to the local maxima.

    Examples
    --------
    >>> import numpy as np
    >>> y = np.array([1, 3, 2, 5, 4, 6])
    >>> find_local_max(y)
    array([1, 3, 5])
    """
    # differenced y
    dy = y[1:] - y[:-1]

    # sign of difference of y
    sign_dy = np.sign(dy)
    sign_dy = np.concatenate((np.array([1]), sign_dy, np.array([-1])))  # add 1 and -1 on the ends for maximum finding

    # find the maxima in y
    extrema = sign_dy[1:] - sign_dy[:-1]  # places where the array dy changes sign
    maxima = np.arange(len(extrema))[extrema < 0]  # places where y has a maximum

    return maxima


@nb.njit(cache=True)
def uphill_local_max(x, y, x_approx):
    """Find the index of the local maximum in `y` uphill from `x_approx`.

    The array `x` must be sorted in ascending order. This function identifies local maxima in the `y` array
    and returns the index of the `x` value corresponding to the local maximum uphill from `x_approx`.

    Parameters
    ----------
    x: numpy.ndarray[Any, dtype[float]]
        1D array of x-values, which must be sorted in ascending order.
    y: numpy.ndarray[Any, dtype[float]]
        1D array of y-values corresponding to the x-values.
    x_approx: numpy.ndarray[Any, dtype[float]]
        The x-value(s) around which to find the nearest local maximum.

    Returns
    -------
    numpy.ndarray[Any, dtype[int]]
        Index of the `x` value(s) corresponding to the nearest local maximum to `x_approx`.

    Examples
    --------
    >>> import numpy as np
    >>> x = np.array([1, 2, 3, 4, 5, 6])
    >>> y = np.array([1, 3, 2, 5, 4, 6])
    >>> x_approx = np.array([2.4, 4.6])
    >>> uphill_local_max(x, y, x_approx)
    array([1, 3])
    """
    # [note] this is find_local_max(y), but since we need sign_dy we do this here
    # differenced y
    dy = y[1:] - y[:-1]

    # find the maxima in y
    sign_dy = np.sign(dy)
    sign_dy = np.concatenate((np.array([1]), sign_dy, np.array([-1])))  # add 1 and -1 on the ends for maximum finding
    extrema = sign_dy[1:] - sign_dy[:-1]  # places where the array dy changes sign
    maxima = np.arange(len(extrema))[extrema < 0]  # places where y has a maximum

    # position of x_approx in x (index is point to the right of x_approx)
    pos_x = np.searchsorted(x, x_approx)  # pos_x can be 0 and len(x) but this is ok for sign_dy

    # do we need to move left or right to get to the maximum
    left_right = sign_dy[pos_x]  # negative: left, positive: right

    # get the maximum for each x_approx (index is maximum to the right of pos_x)
    pos_max = np.searchsorted(maxima, pos_x)
    # take the maximum on the left for negative slopes
    pos_max[left_right == -1] -= 1
    x_max = maxima[pos_max]

    return x_max


@nb.njit(cache=True)
def mask_between(x, locations):
    """Mask out everything except the parts between the given timestamps

    Parameters
    ----------
    x: numpy.ndarray[Any, dtype[float]]
        Time series to be masked.
    locations: numpy.ndarray[Any, dtype[float]]
        Pairs of points in the time series. Everything but the space between points is masked.

    Returns
    -------
    numpy.ndarray[bool]
        Boolean mask that is True between the locations.
    """
    mask = np.zeros(len(x), dtype=np.bool_)
    for loc in locations:
        mask = mask | ((x >= loc[0]) & (x <= loc[-1]))

    return mask


@nb.njit(cache=True)
def mark_gaps(x, min_gap=1.):
    """Mark gaps in a time series.

    Parameters
    ----------
    x: numpy.ndarray[Any, dtype[float]]
        Time series with gaps.
    min_gap: float, optional
        Minimum width for a gap (in time units).

    Returns
    -------
    gaps: numpy.ndarray[Any, dtype[float]]
        Gap timestamps in pairs.
    """
    # mark the gaps
    t_sorted = np.sort(x)
    t_diff = t_sorted[1:] - t_sorted[:-1]  # np.diff(a)
    gaps = (t_diff > min_gap)

    # get the timestamps
    t_left = t_sorted[:-1][gaps]
    t_right = t_sorted[1:][gaps]
    gaps = np.column_stack((t_left, t_right))

    return gaps


@nb.njit(cache=True)
def n_parameters(n_chunks, n_sinusoids, n_harmonics):
    """Return the number of parameters of the model."""
    # equation for number of parameters
    n_param = 2 * n_chunks  # time chunk each with constant and slope
    n_param += int(n_harmonics > 0)  # one period if harmonics present
    n_param += 2 * n_harmonics  # harmonics (sinusoids with constrained frequency)
    n_param += 3 * (n_sinusoids - n_harmonics)  # sinusoids with free frequency

    return n_param


@nb.njit(cache=True)
def correct_for_crowdsap(flux, crowdsap, i_chunks):
    """Correct the flux for flux contribution of a third source
    
    Parameters
    ----------
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    crowdsap: list[float], numpy.ndarray[Any, dtype[float]]
        Light contamination parameter (1-third_light) listed per sector
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    
    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series corrected for
        contaminating light
    
    Notes
    -----
    Uses the parameter CROWDSAP included with some TESS data.
    flux_corrected = (flux - (1 - crowdsap)) / crowdsap
    where all quantities are median-normalised, including the result.
    This corresponds to subtracting a fraction of (1 - crowdsap) of third light
    from the (non-median-normalised) flux measurements.
    """
    cor_flux = np.zeros(len(flux))
    for i, s in enumerate(i_chunks):
        crowd = min(max(0., crowdsap[i]), 1.)  # clip to avoid unphysical output
        cor_flux[s[0]:s[1]] = (flux[s[0]:s[1]] - 1 + crowd) / crowd

    return cor_flux


@nb.njit(cache=True)
def model_crowdsap(flux, crowdsap, i_chunks):
    """Incorporate flux contribution of a third source into the flux

    Parameters
    ----------
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    crowdsap: list[float], numpy.ndarray[Any, dtype[float]]
        Light contamination parameter (1-third_light) listed per sector
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Model of the flux incorporating light contamination

    Notes
    -----
    Does the opposite as correct_for_crowdsap, to be able to model the effect of
    third light to some degree (can only achieve an upper bound on CROWDSAP).
    """
    model = np.zeros(len(flux))
    for i, s in enumerate(i_chunks):
        crowd = min(max(0., crowdsap[i]), 1.)  # clip to avoid unphysical output
        model[s[0]:s[1]] = flux[s[0]:s[1]] * crowd + 1 - crowd

    return model


@nb.njit(cache=True)
def formal_uncertainties_linear(time, residuals, i_chunks):
    """Calculates the corrected uncorrelated (formal) uncertainties for the
    parameters constant and slope.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple
        A tuple containing the following elements:
        sigma_const: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the constant for each sector
        sigma_slope: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the slope for each sector

    Notes
    -----
    Errors in const and slope:
    https://pages.mtu.edu/~fmorriso/cm3215/UncertaintySlopeInterceptOfLeastSquaresFit.pdf
    """
    n_param = 2

    # linear regression uncertainties
    sigma_const = np.zeros(len(i_chunks))
    sigma_slope = np.zeros(len(i_chunks))
    for i, s in enumerate(i_chunks):
        len_t = len(time[s[0]:s[1]])
        n_data = len(residuals[s[0]:s[1]])  # same as len_t, but just for the sake of clarity
        n_dof = n_data - n_param  # degrees of freedom

        # standard deviation of the residuals but per sector
        std = std_unb(residuals[s[0]:s[1]], n_dof)

        # some sums for the uncertainty formulae
        sum_t = 0
        for t in time[s[0]:s[1]]:
            sum_t += t
        ss_xx = 0
        for t in time[s[0]:s[1]]:
            ss_xx += (t - sum_t / len_t)**2
        sigma_const[i] = std * np.sqrt(1 / n_data + (sum_t / len_t)**2 / ss_xx)
        sigma_slope[i] = std / np.sqrt(ss_xx)

    return sigma_const, sigma_slope


@nb.njit(cache=True)
def formal_uncertainties_sinusoid(time, residuals, flux_err, a_n, i_chunks):
    """Calculates the corrected uncorrelated (formal) uncertainties for the frequencies, amplitudes, and phases.

    Harmonics are not taken into account.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple
        A tuple containing the following elements:
        sigma_f: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the frequency for each sinusoid.
        sigma_a: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the amplitude for each sinusoid (these are identical).
        sigma_ph: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the phase for each sinusoid.

    Notes
    -----
    As in Aerts 2021, https://ui.adsabs.harvard.edu/abs/2021RvMP...93a5001A/abstract
    The sigma value in the formulae is approximated by taking the maximum of the
    standard deviation of the residuals and the standard error of the minimum data error.
    """
    n_data = len(residuals)
    n_param = n_parameters(len(i_chunks), len(a_n), 0)  # number of parameters in the model
    n_dof = max(n_data - n_param, 1)  # degrees of freedom

    # calculate the standard deviation of the residuals
    std = std_unb(residuals, n_dof)

    # calculate the standard error based on the smallest data error
    ste = np.median(flux_err) / np.sqrt(n_data)

    # take the maximum of the standard deviation and standard error as sigma N
    sigma_n = max(std, ste)

    # calculate the D factor (square root of the average number of consecutive data points of the same sign)
    positive = (residuals > 0).astype(np.int_)
    indices = np.arange(n_data)
    zero_crossings = indices[1:][np.abs(positive[1:] - positive[:-1]).astype(np.bool_)]
    sss_i = np.concatenate((np.array([0]), zero_crossings, np.array([n_data])))  # same-sign sequence indices
    d_factor = np.sqrt(np.mean(np.diff(sss_i)))

    # uncertainty formulae for sinusoids
    sigma_f = d_factor * sigma_n * np.sqrt(6 / n_data) / (np.pi * a_n * np.ptp(time))
    sigma_a = d_factor * sigma_n * np.sqrt(2 / n_data)
    sigma_ph = d_factor * sigma_n * np.sqrt(2 / n_data) / a_n  # times 2 pi w.r.t. the paper

    # make an array of sigma_a (these are the same)
    sigma_a = np.full(len(a_n), sigma_a)

    return sigma_f, sigma_a, sigma_ph


def formal_uncertainties_harmonic(sigma_f, h_base, h_mult):
    """Calculates the uncertainty in the harmonic frequencies based on the unconstrained frequency
    uncertainties using a simple weighted average.

    The base frequency's uncertainty is calculated as:
    σ_h1 = (∑ n^2/σ_fn^2)^(-1/2)
    The other uncertainties follow as:
    σ_hn = n * σ_h1

    Parameters
    ----------
    sigma_f: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequency for each sinusoid.
    h_base: numpy.ndarray[Any, dtype[int]]
        Indices of the base frequencies in f_n for each of f_n.
    h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each of f_n.

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Uncertainty in the harmonic frequencies.
    """
    # calculate the base frequency uncertainty
    sigma_h = np.zeros(len(sigma_f))
    for i_base in h_base[h_mult == 1]:
        h_series = h_base == i_base  # select one harmonic series at a time
        sigma_h[h_series] = np.sum(h_mult[h_series] ** 2 / sigma_f[h_series] ** 2) ** (-1 / 2)

    # multiply by the harmonic number
    sigma_h *= h_mult

    return sigma_h


@nb.njit(cache=True)
def formal_uncertainties(time, residuals, flux_err, a_n, i_chunks):
    """Calculates the corrected uncorrelated (formal) uncertainties for the extracted
    parameters (constant, slope, frequencies, amplitudes and phases).

    Harmonics are not taken into account.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    residuals: numpy.ndarray[Any, dtype[float]]
        Residual is flux - model
    flux_err: numpy.ndarray[Any, dtype[float]]
        Errors in the measurement values
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sine waves
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple
        A tuple containing the following elements:
        sigma_const: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the constant for each sector
        sigma_slope: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the slope for each sector
        sigma_f: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the frequency for each sine wave
        sigma_a: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the amplitude for each sine wave (these are identical)
        sigma_ph: numpy.ndarray[Any, dtype[float]]
            Uncertainty in the phase for each sine wave

    Notes
    -----
    As in Aerts 2021, https://ui.adsabs.harvard.edu/abs/2021RvMP...93a5001A/abstract
    The sigma value in the formulae is approximated by taking the maximum of the
    standard deviation of the residuals and the standard error of the minimum data error.
    Errors in const and slope:
    https://pages.mtu.edu/~fmorriso/cm3215/UncertaintySlopeInterceptOfLeastSquaresFit.pdf
    """
    n_data = len(residuals)
    n_param = n_parameters(len(i_chunks), len(a_n), 0)  # number of parameters in the model
    n_dof = max(n_data - n_param, 1)  # degrees of freedom

    # calculate the standard deviation of the residuals
    std = std_unb(residuals, n_dof)

    # calculate the standard error based on the smallest data error
    ste = np.median(flux_err) / np.sqrt(n_data)

    # take the maximum of the standard deviation and standard error as sigma N
    sigma_n = max(std, ste)

    # calculate the D factor (square root of the average number of consecutive data points of the same sign)
    positive = (residuals > 0).astype(np.int_)
    indices = np.arange(n_data)
    zero_crossings = indices[1:][np.abs(positive[1:] - positive[:-1]).astype(np.bool_)]
    sss_i = np.concatenate((np.array([0]), zero_crossings, np.array([n_data])))  # same-sign sequence indices
    d_factor = np.sqrt(np.mean(np.diff(sss_i)))

    # uncertainty formulae for sinusoids
    sigma_f = d_factor * sigma_n * np.sqrt(6 / n_data) / (np.pi * a_n * np.ptp(time))
    sigma_a = d_factor * sigma_n * np.sqrt(2 / n_data)
    sigma_ph = d_factor * sigma_n * np.sqrt(2 / n_data) / a_n  # times 2 pi w.r.t. the paper

    # make an array of sigma_a (these are the same)
    sigma_a = np.full(len(a_n), sigma_a)

    # linear regression uncertainties
    sigma_const, sigma_slope = formal_uncertainties_linear(time, residuals, i_chunks)

    return sigma_const, sigma_slope, sigma_f, sigma_a, sigma_ph


def linear_regression_uncertainty_ephem(time, p_orb, sigma_t=1):
    """Calculates the linear regression errors on period and t_zero

    Parameters
    ---------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series.
    p_orb: float
        Orbital period in days.
    sigma_t: float
        Error in the individual time measurements.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        p_err: float
            Error in the period.
        t_err: float
            Error in t_zero.
        p_t_corr: float
            Covariance between the period and t_zero.

    Notes
    -----
    The number of eclipses, computed from the period and
    time base, is taken to be a contiguous set.
    var_matrix:
    [[std[0]**2          , std[0]*std[1]*corr],
     [std[0]*std[1]*corr,           std[1]**2]]
    """
    # number of observed eclipses (technically contiguous)
    n = int(abs(np.ptp(time) // p_orb)) + 1

    # the arrays
    x = np.arange(n, dtype=int)  # 'time' points
    y = np.ones(n, dtype=int)  # 'positive measurement'

    # remove points in gaps
    gaps = mark_gaps(time, min_gap=1.)
    mask = mask_between(x * p_orb, gaps)  # convert x to time domain
    x = x[~mask] - n//2  # also centre the time for minimal correlation
    y = y[~mask]

    # M
    matrix = np.column_stack((x, y))

    # M^-1
    matrix_inv = np.linalg.pinv(matrix)  # inverse (of a general matrix)

    # M^-1 S M^-1^T, S unit matrix times some sigma (no covariance in the data)
    var_matrix = matrix_inv @ matrix_inv.T
    var_matrix = var_matrix * sigma_t ** 2

    # errors in the period and t_zero
    p_err = np.sqrt(var_matrix[0, 0])
    t_err = np.sqrt(var_matrix[1, 1])
    p_t_corr = var_matrix[0, 1] / (t_err * p_err)  # or [1, 0]

    return p_err, t_err, p_t_corr



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This module contains the configuration class.

Code written by: Luc IJspeert
"""
import os
import yaml
import importlib.resources


class Config:
    """A singleton class that manages application configuration settings.

    Attributes
    ----------
    _instance: Config object, None
        The single instance of the Config class, or None.
    _config_path: str
        Path to the configuration file.

    resolution_factor: float
        Number that multiplies the resolution (1/T).

    Methods
    -------
    __new__(cls)
        Ensures only one instance of the Config class is created.
    _load_config(cls)
        Loads and validates the configuration from a file.
    _validate_config(cls, config_data)
        Validates the loaded configuration data.
    _initialize_default_settings(cls)
        Initializes default settings if the configuration file is not found or invalid.
    """
    _instance = None
    _config_path = os.path.join(os.path.dirname(__file__), 'config.yaml')

    # default settings below
    # General settings
    verbose: bool = False
    stop_at_stage: int = 0

    # Extraction settings
    select_next: str = 'hybrid'
    optimise_step: bool = True
    replace_step: bool = True
    stop_criterion: str = 'bic'
    bic_thr: float = 2.
    snr_thr: float = -1.
    nyquist_factor: float = 1.
    resolution_factor: float = 1.5
    window_width: float = 1.

    # Data and file settings
    overwrite: bool = False
    data_dir: str = ''
    save_dir: str = ''
    save_ascii: bool = False

    # Tabulated files settings
    cn_time: str = 'time'
    cn_flux: str = 'flux'
    cn_flux_err: str = 'flux_err'

    # Fits files settings
    cf_time: str = 'TIME'
    cf_flux: str = 'SAP_FLUX'
    cf_flux_err: str = 'SAP_FLUX_ERR'
    cf_quality: str = 'QUALITY'
    apply_q_flags: bool = True
    halve_chunks: bool = False

    # GUI settings
    dark_mode: bool = False
    h_size_frac: float = 0.8
    v_size_frac: float = 0.8

    def __new__(cls):
        """Ensures that only one instance of the Config class is created.

        Returns
        -------
        Config
            The single instance of the Config class.
        """
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
            cls._load_config(cls._instance)

        return cls._instance

    @classmethod
    def _validate_config(cls, config_data):
        """Validates the loaded configuration data.

        Parameters
        ----------
        config_data: dict
            The loaded configuration data.

        Raises
        ------
        ValueError
            If the configuration data is invalid.
        """
        # make sets out of the keywords
        expected_keys = set(key for key in dir(cls) if not callable(getattr(cls, key)) and not key.startswith('_'))
        config_keys = set(config_data.keys())

        # test whether config_keys is missing items
        missing_keys = expected_keys - config_keys
        if len(missing_keys) != 0:
            raise ValueError(f"Missing keys in configuration file: {missing_keys}")

        # test whether config_keys has excess items
        excess_keys = config_keys - expected_keys
        if len(excess_keys) != 0:
            raise ValueError(f"Excess keys in configuration file: {excess_keys}")

        return None

    def _load_config(self):
        """Loads and validates the configuration from a file.

        Notes
        -----
        - `FileNotFoundError`: Catches the error if the configuration file is not found.
        - `YAMLError`: Catches errors related to parsing YAML files.
        - `ValueError`: Catches errors related to the validation of the configuration.
        In these cases, the default settings are loaded.
        """
        # try to open the config file
        try:
            with open(self._config_path, 'r') as file:
                config_data = yaml.safe_load(file)
                self._validate_config(config_data)
                for key, value in config_data.items():
                    setattr(self._instance, key, value)

        except FileNotFoundError:
            print(f"Configuration file {self._config_path} not found. Using default settings.")

        except yaml.YAMLError as e:
            print(f"Error parsing YAML from {self._config_path}: {e}. Using default settings.")

        except ValueError as e:
            print(f"Error validating configuration from {self._config_path}: {e}. Your config file may be out of date. "
                  f"Using default settings.")

        return None

    def update_from_file(self, new_config_path):
        """Updates the settings with a user defined configuration file.

        Parameters
        ----------
        new_config_path: str
            Path to a valid configuration file.
        """
        self._config_path = new_config_path
        self._load_config()

        return None

    def update_from_dict(self, settings):
        """Updates the settings with user defined keyword arguments.

        Parameters
        ----------
        settings: dict
            Configuration settings.
        """
        # remember invalid items
        invalid = dict()

        # set the valid attributes
        for key, value in settings.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                invalid[key] = value

        # Print a message about items that were invalid
        if len(invalid) > 0:
            print(f"Invalid items that were not updated: {invalid}")

        return None

    def save_to_file(self, new_config_path):
        """Save the configuration to a YAML file manually.

        Parameters
        ----------
        new_config_path: str
            Path to a valid configuration file.
        """
        # make sure the file has yaml extension
        if not (new_config_path.endswith('.yaml') or new_config_path.endswith('.yml')):
            new_config_path += '.yaml'

        # if save_dir is usr dir, revert to '' for saving (this is specifically for GUI behaviour)
        save_dir_value = self.save_dir
        if self.save_dir == os.path.expanduser('~'):
            save_dir_value = ''

        line_width = 120

        with open(new_config_path, 'w') as file:
            file.write("#" * line_width + "\n")
            file.write(fill_header_str("Star Shine settings file", line_width, fill_value='-', end='\n'))
            file.write("#" * line_width + "\n")

            file.write(fill_header_str("General settings", line_width, fill_value='-', end='\n'))

            desc = "Print information during runtime"
            file.write(config_item_description("verbose", self.verbose, desc))

            desc = "Run the analysis up to and including this stage; 0 means all stages are run"
            file.write(config_item_description("stop_at_stage", self.stop_at_stage, desc))

            file.write(fill_header_str("Extraction settings", line_width, fill_value='-', end='\n'))

            desc = ("Select the next frequency in iterative extraction based on 'amp', 'snr', "
                    "or 'hybrid' (first amp then snr)")
            file.write(config_item_description("select_next", self.select_next, desc))

            desc = "Optimise with a non-linear multi-sinusoid fit at every step (T) or only at the end (F)"
            file.write(config_item_description("optimise_step", self.optimise_step, desc))

            desc = ("Attempt to replace closely spaced sinusoids by one sinusoid "
                    "at every step (T) or only at the end (F)")
            file.write(config_item_description("replace_step", self.replace_step, desc))

            desc = "Stop criterion for the iterative extraction of sinusoids will be based on 'bic', or 'snr'"
            file.write(config_item_description("stop_criterion", self.stop_criterion, desc))

            desc = "Delta-BIC threshold for the acceptance of sinusoids"
            file.write(config_item_description("bic_thr", self.bic_thr, desc))

            desc = "Signal-to-noise threshold for the acceptance of sinusoids, uses a built-in method if set to -1"
            file.write(config_item_description("snr_thr", self.snr_thr, desc))

            desc = "The simple Nyquist frequency approximation (1/(2 delta_t_min)) is multiplied by this factor"
            file.write(config_item_description("nyquist_factor", self.nyquist_factor, desc))

            desc = "The frequency resolution (1/T) is multiplied by this factor"
            file.write(config_item_description("resolution_factor", self.resolution_factor, desc))

            desc = "Periodogram spectral noise is calculated over this window width"
            file.write(config_item_description("window_width", self.window_width, desc))

            file.write("#" * line_width + "\n")
            file.write(fill_header_str("Data and File settings", line_width, fill_value='-', end='\n'))

            desc = "Overwrite existing result files"
            file.write(config_item_description("overwrite", self.overwrite, desc))

            desc = "Root directory where the data files to be analysed are located; if empty will use current dir"
            file.write(config_item_description("data_dir", self.data_dir, desc))

            desc = "Root directory where analysis results will be saved; if empty will use current dir"
            file.write(config_item_description("save_dir", save_dir_value, desc))

            desc = "Save ascii variants of the HDF5 result files"
            file.write(config_item_description("save_ascii", self.save_ascii, desc))

            file.write(fill_header_str("Tabulated File settings", line_width, fill_value='-', end='\n'))

            desc = "Column name for the time stamps"
            file.write(config_item_description("cn_time", self.cn_time, desc))

            desc = "Column name for the flux measurements"
            file.write(config_item_description("cn_flux", self.cn_flux, desc))

            desc = "Column name for the flux measurement errors"
            file.write(config_item_description("cn_flux_err", self.cn_flux_err, desc))

            file.write(fill_header_str("FITS File settings", line_width, fill_value='-', end='\n'))

            desc = "Column name for the time stamps"
            file.write(config_item_description("cf_time", self.cf_time, desc))

            desc = "Column name for the flux [examples: SAP_FLUX, PDCSAP_FLUX, KSPSAP_FLUX]"
            file.write(config_item_description("cf_flux", self.cf_flux, desc))

            desc = "Column name for the flux errors [examples: SAP_FLUX_ERR, PDCSAP_FLUX_ERR, KSPSAP_FLUX_ERR]"
            file.write(config_item_description("cf_flux_err", self.cf_flux_err, desc))

            desc = "Column name for the flux quality flags"
            file.write(config_item_description("cf_quality", self.cf_quality, desc))

            desc = "Apply the quality flags supplied by the data source"
            file.write(config_item_description("apply_q_flags", self.apply_q_flags, desc))

            desc = "Cut the time chunks in half (TESS data often has a discontinuity mid-sector)"
            file.write(config_item_description("halve_chunks", self.halve_chunks, desc))

            file.write("#" * line_width + "\n")
            file.write(fill_header_str("GUI settings", line_width, fill_value='-', end='\n'))

            desc = "Dark mode. [WIP]"
            file.write(config_item_description("dark_mode", self.dark_mode, desc))

            desc = "Horizontal window size as a fraction of the screen width"
            file.write(config_item_description("h_size_frac", self.h_size_frac, desc))

            desc = "Vertical window size as a fraction of the screen height"
            file.write(config_item_description("v_size_frac", self.v_size_frac, desc))

            file.write("#" * line_width + "\n")

        return None


def fill_header_str(header, line_width, fill_value='-', end='\n'):
    """Fill a header string on eiter side with hyphens.

    Parameters
    ----------
    header: str
        Text to put in the centre of the header.
    line_width: int
        Width of the final string, excluding `end`.
    fill_value: str
        Value used to fill up the line.
    end: str
        Something to end the line.

    Returns
    -------
    str
        Filled header string.
    """
    # add spaces to the header
    header = " " + header + " "

    # calculate the space to fill on either side
    left = (line_width - len(header)) // 2
    right = line_width - left - len(header)

    header = "# " + fill_value * (left - 2) + header + fill_value * right + end

    return header


def config_item_description(item, value, description):
    """Return a description string for a setting in the config file.

    Parameters
    ----------
    item: str
        The name of the configuration item.
    value: any
        The value of the configuration item.
    description: str
        A description of what the configuration item does.

    Returns
    -------
    str
        A formatted string that includes the item, its value, and a description.
    """
    # start with standardised bit
    item_desc = f"# {item} description:\n"

    # add the description
    item_desc += f"# {description}\n"

    # if item is already a string, add ''
    if type(value) is str:
        value = "'" + value + "'"

    # finally add the item
    item_desc += f"{item}: {value}\n\n"  # empty line at the end

    return item_desc


def get_config():
    """Use this function to get the configuration

    Returns
    -------
    Config
        The singleton instance of Config.
    """
    return Config()


def get_config_path():
    """Get the path to the configuration file

    Returns
    -------
    str
        Path to the config file
    """
    # Use importlib.resources to find the path
    config_path = str(importlib.resources.files('star_shine.config').joinpath('config.yaml'))

    return config_path



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains classes for handling the models for the time series. Includes a piece-wise linear model,
sum of sinusoids, and harmonics.

Code written by: Luc IJspeert
"""
import numpy as np
import numba as nb

from star_shine.core import frequency_sets as frs, periodogram as pdg, utility as ut


@nb.njit(cache=True, parallel=True)
def linear_curve(time, const, slope, i_chunks, t_shift=True):
    """Returns a piece-wise linear curve for the given time points
    with slopes and y-intercepts.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve
    slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
    t_shift: bool
        Mean center the time axis

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        The model time series of a (set of) straight line(s)

    Notes
    -----
    Assumes the constants and slopes are determined with respect to the sector mean time as zero point.
    """
    curve = np.zeros(len(time))
    for i in nb.prange(len(const)):
        s = i_chunks[i]
        if t_shift:
            t_sector_mean = np.mean(time[s[0]:s[1]])
        else:
            t_sector_mean = 0

        curve[s[0]:s[1]] = const[i] + slope[i] * (time[s[0]:s[1]] - t_sector_mean)

    return curve


@nb.njit(cache=True, parallel=True)
def linear_pars(time, flux, i_chunks):
    """Calculate the slopes and y-intercepts of a linear trend with the MLE.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    flux: numpy.ndarray[Any, dtype[float]]
        Measurement values of the time series
    i_chunks: numpy.ndarray[Any, dtype[int]]
        Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
        the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

    Returns
    -------
    tuple
        A tuple containing the following elements:
        y_inter: numpy.ndarray[Any, dtype[float]]
            The y-intercepts of a piece-wise linear curve
        slope: numpy.ndarray[Any, dtype[float]]
            The slopes of a piece-wise linear curve

    Notes
    -----
    Source: https://towardsdatascience.com/linear-regression-91eeae7d6a2e
    Determines the constants and slopes with respect to the sector mean time as zero point to avoid correlations.
    """
    y_inter = np.zeros(len(i_chunks))
    slope = np.zeros(len(i_chunks))

    for i in nb.prange(len(i_chunks)):
        s = i_chunks[i]

        # mean and mean subtracted quantities
        x_m = np.mean(time[s[0]:s[1]])
        x_ms = (time[s[0]:s[1]] - x_m)
        y_m = np.mean(flux[s[0]:s[1]])
        y_ms = (flux[s[0]:s[1]] - y_m)

        # sums
        s_xx = np.sum(x_ms ** 2)
        s_xy = np.sum(x_ms * y_ms)

        # parameters
        slope[i] = s_xy / s_xx
        # y_inter[i] = y_m - slope[i] * x_m  # original non-mean-centered formula
        y_inter[i] = y_m  # mean-centered value

    return y_inter, slope


@nb.njit(cache=True)
def sum_sines_st(time, f_n, a_n, ph_n, t_shift=True):
    """A sum of sinusoids at times t, given the frequencies, amplitudes and phases.

    Single threaded version. Better for one to a few sinusoids.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sinusoids
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sinusoids
    t_shift: bool
        Mean center the time axis

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Model time series of a sum of sinusoids. Varies around 0.

    Notes
    -----
    Assumes the phases are determined with respect to the mean time as zero point by default.
    """
    if t_shift:
        mean_t = np.mean(time)
    else:
        mean_t = 0

    model_sines = np.zeros(len(time))
    for i in range(len(f_n)):
        model_sines += a_n[i] * np.sin((2 * np.pi * f_n[i] * (time - mean_t)) + ph_n[i])

    return model_sines


@nb.njit(cache=True, parallel=True)
def sum_sines(time, f_n, a_n, ph_n, t_shift=True):
    """A sum of sinusoids at times t, given the frequencies, amplitudes and phases.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sinusoids
    a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids
    ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sinusoids
    t_shift: bool
        Mean center the time axis

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Model time series of a sum of sinusoids. Varies around 0.

    Notes
    -----
    Assumes the phases are determined with respect to the mean time as zero point by default.
    """
    if t_shift:
        mean_t = np.mean(time)
    else:
        mean_t = 0

    model_sines = np.zeros(len(time))
    for i in nb.prange(len(f_n)):
        model_sines += a_n[i] * np.sin((2 * np.pi * f_n[i] * (time - mean_t)) + ph_n[i])

    return model_sines


@nb.njit(cache=True, parallel=True)
def sum_sines_deriv(time, f_n, a_n, ph_n, deriv=1, t_shift=True):
    """The time derivative of a sum of sinusoids at times t.

    Parameters
    ----------
    time: numpy.ndarray[Any, dtype[float]]
        Timestamps of the time series
    f_n: list[float], numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sinusoids
    a_n: list[float], numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids
    ph_n: list[float], numpy.ndarray[Any, dtype[float]]
        The phases of a number of sinusoids
    deriv: int
        Number of time derivatives taken (>= 1)
    t_shift: bool
        Mean center the time axis

    Returns
    -------
    numpy.ndarray[Any, dtype[float]]
        Model time series of a sum of sinusoid derivatives. Varies around 0.

    Notes
    -----
    Assumes the phases are determined with respect to the mean time as zero point by default.
    """
    if t_shift:
        mean_t = np.mean(time)
    else:
        mean_t = 0

    model_sines = np.zeros(len(time))
    mod_2 = deriv % 2
    mod_4 = deriv % 4
    ph_cos = (np.pi / 2) * mod_2  # alternate between cosine and sine
    sign = (-1) ** ((mod_4 - mod_2) // 2)  # (1, -1, -1, 1, 1, -1, -1... for deriv=1, 2, 3...)

    for i in nb.prange(len(f_n)):
        for j in range(len(time)):
            model_sines[j] += (sign * (2 * np.pi * f_n[i]) ** deriv * a_n[i] *
                               np.sin((2 * np.pi * f_n[i] * (time[j] - mean_t)) + ph_n[i] + ph_cos))

    return model_sines


class LinearModel:
    """This class handles the linear model.

    Attributes
    ----------
    _const: numpy.ndarray[Any, dtype[float]]
        The y-intercepts of a piece-wise linear curve.
    _slope: numpy.ndarray[Any, dtype[float]]
        The slopes of a piece-wise linear curve.
    _const_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the constant for each time chunk.
    _slope_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the slope for each time chunk.
    _linear_model: numpy.ndarray[Any, dtype[float]]
        Time series model of the piece-wise linear curve.
    """

    def __init__(self, n_time):
        """Initialises the Result object.

        Parameters
        ----------
        n_time: int
            Number of points in the time series.
        """
        # linear model parameters
        self._const = np.zeros((0,))  # y-intercepts
        self._slope = np.zeros((0,))  # slopes

        # linear parameter uncertainties
        self._const_err = np.zeros((0,))
        self._slope_err = np.zeros((0,))

        # number of time chunks
        self.n_chunks = 0

        # current model
        self._linear_model = np.zeros((n_time,))

    @property
    def const(self):
        """Get the current model constants.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The y-intercepts of a piece-wise linear curve.
        """
        return self._const.copy()

    @property
    def slope(self):
        """Get the current model slopes.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The slopes of a piece-wise linear curve.
        """
        return self._slope.copy()

    @property
    def const_err(self):
        """Get the current model constants.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The y-intercepts of a piece-wise linear curve.
        """
        return self._const_err.copy()

    @property
    def slope_err(self):
        """Get the current model slopes.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The slopes of a piece-wise linear curve.
        """
        return self._slope_err.copy()

    @property
    def linear_model(self):
        """Get the current linear model.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Time series model of the piece-wise linear curve.
        """
        return self._linear_model.copy()

    @property
    def n_param(self):
        """Get the number of parameters of the model."""
        return 2 * self.n_chunks

    def update_n(self):
        """Update the current numbers of sinusoids, harmonics, and base frequencies."""
        self.n_chunks = len(self._const)

        return None

    def get_linear_parameters(self):
        """Get a copy of the current linear parameters.

        Returns
        -------
        tuple
            Consisting of two numpy.ndarray[Any, dtype[float]] for const, slope.
        """
        return self.const, self.slope

    def calc_linear_model(self, time, i_chunks):
        """Calculate the current linear model.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Time series model of the piece-wise linear curve.
        """
        return linear_curve(time, self.const, self.slope, i_chunks)

    def set_linear_model(self, time, const_new, slope_new, i_chunks):
        """Set the linear model according to the new parameters.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        const_new: numpy.ndarray[Any, dtype[float]]
            New y-intercepts.
        slope_new: numpy.ndarray[Any, dtype[float]]
            New slopes.
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        # make the new model
        self._linear_model = linear_curve(time, const_new, slope_new, i_chunks)

        # set the parameters
        self._const = const_new
        self._slope = slope_new

        # set the numbers
        self.update_n()

        return

    def update_linear_model(self, time, residual, i_chunks):
        """Update the linear model using the residual flux.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        residual: numpy.ndarray[Any, dtype[float]]
            Residual flux (flux minus model).
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        # get new parameters
        const_new, slope_new = linear_pars(time, residual, i_chunks)

        # set the new parameters and model
        self.set_linear_model(time, const_new, slope_new, i_chunks)

        return

    def update_linear_uncertainties(self, time, residual, i_chunks):
        """Update the linear model parameter errors using the residual flux.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        residual: numpy.ndarray[Any, dtype[float]]
            Residual flux (flux minus model).
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        self._const_err, self._slope_err = ut.formal_uncertainties_linear(time, residual, i_chunks)

        return


class SinusoidModel:
    """This class handles the sinusoid model.

    Attributes
    ----------
    _f_n: numpy.ndarray[Any, dtype[float]]
        The frequencies of a number of sinusoids.
    _a_n: numpy.ndarray[Any, dtype[float]]
        The amplitudes of a number of sinusoids.
    _ph_n: numpy.ndarray[Any, dtype[float]]
        The phases of a number of sinusoids.
    _include: numpy.ndarray[Any, dtype[bool]]
        Include state of the sinusoids in the model.
    _harmonics: numpy.ndarray[Any, dtype[bool]]
        Boolean mask indicating harmonics in _f_n. False indicates a free sinusoid.
    _h_base: numpy.ndarray[Any, dtype[int]]
        Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
    _h_mult: numpy.ndarray[Any, dtype[int]]
        Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
    _combinations: dict[int, list[int]]
        Indices of the child frequencies in _f_n, and the corresponding parents.
    _combination_n: dict[int, numpy.ndarray[Any, dtype[int]]]
        Indices of the child frequencies in _f_n, and the corresponding parent multiplier.
    _f_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequency for each sinusoid.
    _a_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the amplitude for each sinusoid (these are identical).
    _ph_n_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the phase for each sinusoid.
    _f_h_err: numpy.ndarray[Any, dtype[float]]
        Uncertainty in the frequency for each harmonic sinusoid.
    _passing_sigma: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the sigma check.
    _passing_snr: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the signal-to-noise check.
    _passing_harmonic: numpy.ndarray[Any, dtype[bool]]
        Sinusoids that passed the harmonic check.
    n_sin: int
        Number of sinusoids (including harmonics).
    n_harm: int
        Number of harmonic sinusoids.
    n_base: int
        Number of harmonic series/base frequencies.
    _sinusoid_model: numpy.ndarray[Any, dtype[float]]
        Current time series model of the sinusoids.
    """

    def __init__(self, n_time):
        """Initialises the Result object.

        Parameters
        ----------
        n_time: int
            Number of points in the time series.
        """
        # sinusoid model parameters
        self._f_n = np.zeros((0,))  # frequencies
        self._a_n = np.zeros((0,))  # amplitudes
        self._ph_n = np.zeros((0,))  # phases

        # for consistency of indices in the removal of sinusoids
        self._include = np.zeros((0,), dtype=bool)

        # harmonic model parameters
        self._harmonics = np.zeros((0,), dtype=bool)
        self._h_base = -np.ones((0,), dtype=int)
        self._h_mult = np.zeros((0,), dtype=int)

        # combination model parameters [wip: logic not implemented yet]
        self._combinations = dict()
        self._combination_n = dict()

        # sinusoid parameter uncertainties
        self._f_n_err = np.zeros((0,))
        self._a_n_err = np.zeros((0,))
        self._ph_n_err = np.zeros((0,))

        # harmonic parameter uncertainties
        self._f_h_err = np.zeros((0,))

        # passing criteria
        self._passing_sigma = np.zeros((0,), dtype=bool)
        self._passing_snr = np.zeros((0,), dtype=bool)
        self._passing_harmonic = np.zeros((0,), dtype=bool)

        # number of sinusoids
        self.n_sin = 0  # number of sinusoids (including harm)
        self.n_harm = 0  # number of harmonics
        self.n_base = 0  # number of harmonic series

        # current model
        self._sinusoid_model = np.zeros((n_time,))

    @property
    def f_n(self):
        """Get a copy of the current model frequencies (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The frequencies of a number of sinusoids.
        """
        return self._f_n.copy()

    @property
    def a_n(self):
        """Get a copy of the current model amplitudes (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The amplitudes of a number of sinusoids.
        """
        return self._a_n.copy()

    @property
    def ph_n(self):
        """Get a copy of the current model phases (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The phases of a number of sinusoids.
        """
        return self._ph_n.copy()

    @property
    def include(self):
        """Get a copy of the current include state of the sinusoids.

        Returns
        -------
        numpy.ndarray[Any, dtype[bool]]
            Include state of the sinusoids in the model.
        """
        return self._include.copy()

    @property
    def has_excludes(self):
        """Check if there are any not included sinusoids."""
        return np.any(~self._include)

    @property
    def harmonics(self):
        """Get a copy of the current model harmonics (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[bool]]
            Boolean mask indicating harmonics in _f_n. False indicates a free sinusoid.
        """
        return self._harmonics.copy()

    @property
    def h_base(self):
        """Get a copy of the current model harmonic base frequency indices (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[int]]
            Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
        """
        return self._h_base.copy()

    @property
    def h_mult(self):
        """Get a copy of the current model harmonic multiplier numbers (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[int]]
            Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
        """
        return self._h_mult.copy()

    @property
    def f_base(self):
        """Get the current model base frequencies (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The base frequencies of the harmonic model.
        """
        return self._f_n[self._h_mult == 1]

    @property
    def f_n_err(self):
        """Get a copy of the model frequency errors (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The errors in the frequencies of a number of sinusoids.

        Notes
        -----
        These need to be updated manually (with update_sinusoid_uncertainties).
        """
        return self._f_n_err.copy()

    @property
    def a_n_err(self):
        """Get a copy of the model amplitude errors (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The errors in the amplitudes of a number of sinusoids.

        Notes
        -----
        These need to be updated manually (with update_sinusoid_uncertainties).
        """
        return self._a_n_err.copy()

    @property
    def ph_n_err(self):
        """Get a copy of the model phase errors (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The errors in the phases of a number of sinusoids.

        Notes
        -----
        These need to be updated manually (with update_sinusoid_uncertainties).
        """
        return self._ph_n_err.copy()

    @property
    def f_h_err(self):
        """Get a copy of the model harmonic frequency errors (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            The errors in the frequencies of a number of (harmonic) sinusoids.

        Notes
        -----
        These need to be updated manually (with update_sinusoid_uncertainties_harmonic).
        """
        return self._f_h_err.copy()

    @property
    def passing_sigma(self):
        """Get a copy of the passing sigma mask (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[bool]]
            Sinusoids that passed the sigma check.

        Notes
        -----
        These need to be updated manually (with update_sinusoid_passing_sigma).
        """
        return self._passing_sigma.copy()

    @property
    def passing_snr(self):
        """Get a copy of the passing snr mask (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[bool]]
            Sinusoids that passed the signal-to-noise check.
        """
        return self._passing_snr.copy()

    @property
    def passing_harmonic(self):
        """Get a copy of the passing harmonic mask (disregarding include).

        Returns
        -------
        numpy.ndarray[Any, dtype[bool]]
            Sinusoids that passed the harmonic check.
        """
        return self._passing_harmonic.copy()

    @property
    def sinusoid_model(self):
        """Get a copy of the current sinusoid model.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Time series model of the sinusoids.
        """
        return self._sinusoid_model.copy()

    @property
    def n_param(self):
        """Get the number of parameters of the model."""
        return self.n_base + 2 * self.n_harm + 3 * (self.n_sin - self.n_harm)

    def update_n(self):
        """Update the current numbers of sinusoids, harmonics, and base frequencies."""
        self.n_sin = len(self._f_n[self._include])
        self.n_harm = len(self._f_n[self._include][self._harmonics[self._include]])
        self.n_base = len(np.unique(self._h_base[(self._h_base != -1) & self._include]))

        return None

    def get_sinusoid_parameters(self, exclude=True):
        """Get a copy of the current sinusoid parameters.

        Parameters
        ----------
        exclude: bool
            Exclude the sinusoids intended for removal.

        Returns
        -------
        tuple
            Consisting of three numpy.ndarray[Any, dtype[float]] for f_n, a_n, ph_n.
        """
        if exclude:
            f_n, a_n, ph_n = self._f_n[self._include], self._a_n[self._include], self._ph_n[self._include]
        else:
            f_n, a_n, ph_n = self._f_n, self._a_n, self._ph_n

        return f_n, a_n, ph_n

    def get_harmonic_parameters(self, exclude=True):
        """Get a copy of the current harmonic parameters.

        Parameters
        ----------
        exclude: bool
            Exclude the sinusoids intended for removal.

        Returns
        -------
        tuple
            Consisting of three numpy.ndarray[Any, dtype[float]] for f_n, a_n, ph_n.
        """
        if exclude:
            harmonics = self._harmonics[self._include]
            h_base, h_mult = self._h_base[self._include], self._h_mult[self._include]
        else:
            harmonics, h_base, h_mult = self._harmonics, self._h_base, self._h_mult

        return harmonics, h_base, h_mult

    def get_h_base_map(self):
        """Get the indices of each base harmonic frequency and a map to recreate h_base[harmonics].

        Returns
        -------
        tuple
            numpy.ndarray[Any, dtype[int]]
                Indices of the base harmonic frequencies in _f_n.
            numpy.ndarray[Any, dtype[int]]
                Map of harmonics to base frequencies.
        """
        return np.unique(self.h_base[self._harmonics], return_inverse=True)

    def get_f_index(self, f):
        """Get the index in f_n of a given frequency.

        Parameters
        ----------
        f: float
            Frequency to get the index of.

        Returns
        -------
        int
            Index of f in f_n.
        """
        return np.abs(self._f_n - f).argmin()

    def calc_sinusoid_model(self, time, indices=None):
        """Calculate the current sinusoid model (disregarding include).

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        indices: numpy.ndarray[Any, dtype[int]]
            Indices for the sinusoids to include.

        Returns
        -------
        numpy.ndarray[Any, dtype[float]]
            Time series model of the sinusoids.
        """
        if indices is None:
            model = sum_sines(time, self.f_n, self.a_n, self.ph_n)
        else:
            model = sum_sines(time, self.f_n[indices], self.a_n[indices], self.ph_n[indices])

        return model

    def _check_removed_h_base(self, indices):
        """If a harmonic base frequency was removed, remove the whole harmonic series.

        Parameters
        ----------
        indices: numpy.ndarray[Any, dtype[int]]
            Indices for the sinusoids that were removed.
        """
        if np.any(self._harmonics):
            for i in np.arange(len(self._f_n))[self._h_mult == 1]:
                if i in indices:
                    series_mask = self._h_base == i
                    self._harmonics[series_mask] = False
                    self._h_base[series_mask] = -1
                    self._h_mult[series_mask] = 0

        return None

    def enforce_harmonics(self):
        """Enforce harmonic relations in the sinusoids.

        This will set all harmonic frequencies to the integer multiple of the base frequency that they are
        assigned to. Optionally performs a linear fit of the amplitudes and phases [wip].
        """
        if np.any(self._harmonics):
            self._f_n[self._harmonics] = self._f_n[self._h_base[self._harmonics]] * self._h_mult[self._harmonics]

        return None

    def set_sinusoids(self, time, f_n_new, a_n_new, ph_n_new, h_base_new=None, h_mult_new=None):
        """Set the current sinusoid model with the new parameters.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        f_n_new: numpy.ndarray[Any, dtype[float]]
            The frequencies of a number of sinusoids.
        a_n_new: numpy.ndarray[Any, dtype[float]]
            The amplitudes of a number of sinusoids.
        ph_n_new: numpy.ndarray[Any, dtype[float]]
            The phases of a number of sinusoids.
        h_base_new: numpy.ndarray[Any, dtype[int]], optional
            Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
        h_mult_new: numpy.ndarray[Any, dtype[int]], optional
            Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
        """
        # ensure 1d
        f_n_new = np.atleast_1d(f_n_new)
        a_n_new = np.atleast_1d(a_n_new)
        ph_n_new = np.atleast_1d(ph_n_new)

        n_new = len(f_n_new)

        # ensure harmonic input ok
        if h_base_new is None:
            h_base_new = -np.ones(n_new, dtype=int)
        h_base_new = np.atleast_1d(h_base_new)
        if h_mult_new is None:
            h_mult_new = np.zeros(n_new, dtype=int)
        h_mult_new = np.atleast_1d(h_mult_new)
        harmonics_new = h_base_new != -1

        # make the new model
        self._sinusoid_model = sum_sines(time, f_n_new, a_n_new, ph_n_new)

        # update the sinusoid parameters
        self._f_n = f_n_new
        self._a_n = a_n_new
        self._ph_n = ph_n_new
        self._include = np.ones(n_new, dtype=bool)

        # update harmonic parameters
        self._harmonics = harmonics_new
        self._h_base = h_base_new
        self._h_mult = h_mult_new

        # update numbers
        self.update_n()

        return None

    def add_sinusoids(self, time, f_n_new, a_n_new, ph_n_new, h_base_new=None, h_mult_new=None):
        """Add the sinusoids to the list.

        Meant for adding a limited number of sinusoids, less efficient for large numbers.
        For that case, see set_sinusoids.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        f_n_new: numpy.ndarray[Any, dtype[float]]
            The frequencies of a number of sinusoids.
        a_n_new: numpy.ndarray[Any, dtype[float]]
            The amplitudes of a number of sinusoids.
        ph_n_new: numpy.ndarray[Any, dtype[float]]
            The phases of a number of sinusoids.
        h_base_new: numpy.ndarray[Any, dtype[int]], optional
            Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
        h_mult_new: numpy.ndarray[Any, dtype[int]], optional
            Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
        """
        # ensure 1d
        f_n_new = np.atleast_1d(f_n_new)
        a_n_new = np.atleast_1d(a_n_new)
        ph_n_new = np.atleast_1d(ph_n_new)

        n_new = len(f_n_new)

        # ensure harmonic input ok
        if h_base_new is None:
            h_base_new = -np.ones(n_new, dtype=int)
        h_base_new = np.atleast_1d(h_base_new)
        if h_mult_new is None:
            h_mult_new = np.zeros(n_new, dtype=int)
        h_mult_new = np.atleast_1d(h_mult_new)
        harmonics_new = h_base_new != -1

        # make the new model
        new_model = sum_sines_st(time, f_n_new, a_n_new, ph_n_new)

        # update the model
        self._sinusoid_model += new_model

        # update the sinusoid parameters
        self._f_n = np.append(self._f_n, f_n_new)
        self._a_n = np.append(self._a_n, a_n_new)
        self._ph_n = np.append(self._ph_n, ph_n_new)
        self._include = np.append(self._include, np.ones(n_new, dtype=bool))

        # update harmonic parameters
        self._harmonics = np.append(self._harmonics, harmonics_new)
        self._h_base = np.append(self._h_base, h_base_new)
        self._h_mult = np.append(self._h_mult, h_mult_new)

        # update numbers
        self.update_n()

        return None

    def update_sinusoids(self, time, f_n_new, a_n_new, ph_n_new, indices, h_base_new=None, h_mult_new=None):
        """Update the current sinusoid model with changes at the given indices.

        Meant for updating a limited number of sinusoids, less efficient for large numbers.
        For that case, see set_sinusoids.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        f_n_new: numpy.ndarray[Any, dtype[float]]
            The frequencies of a number of sinusoids.
        a_n_new: numpy.ndarray[Any, dtype[float]]
            The amplitudes of a number of sinusoids.
        ph_n_new: numpy.ndarray[Any, dtype[float]]
            The phases of a number of sinusoids.
        indices: numpy.ndarray[Any, dtype[int]]
            Indices for the sinusoids to update.
        h_base_new: numpy.ndarray[Any, dtype[int]], optional
            Indices of the base frequencies in _f_n for each of _f_n. -1 indicates a free sinusoid.
        h_mult_new: numpy.ndarray[Any, dtype[int]], optional
            Harmonic multiplier of the base frequency for each of _f_n. 0 indicates a free sinusoid.
        """
        # ensure 1d
        f_n_new = np.atleast_1d(f_n_new)
        a_n_new = np.atleast_1d(a_n_new)
        ph_n_new = np.atleast_1d(ph_n_new)
        indices = np.atleast_1d(indices)

        n_new = len(f_n_new)

        # ensure harmonic input ok
        if h_base_new is None:
            h_base_new = -np.ones(n_new, dtype=int)
        h_base_new = np.atleast_1d(h_base_new)
        if h_mult_new is None:
            h_mult_new = np.zeros(n_new, dtype=int)
        h_mult_new = np.atleast_1d(h_mult_new)
        harmonics_new = h_base_new != -1

        # if a full list of sinusoid parameters is given, pick out the ones at the indices
        if len(f_n_new) == len(self._f_n):
            f_n_new, a_n_new, ph_n_new = f_n_new[indices], a_n_new[indices], ph_n_new[indices]
            harmonics_new, h_base_new, h_mult_new = harmonics_new[indices], h_base_new[indices], h_mult_new[indices]

        # get a list of indices that are currently included in the model
        i_include = indices[self._include[indices]]

        # get the current model at the indices
        cur_model_i = sum_sines_st(time, self._f_n[i_include], self._a_n[i_include], self._ph_n[i_include])

        # make the new model
        new_model = sum_sines_st(time, f_n_new, a_n_new, ph_n_new)

        # update the model
        self._sinusoid_model = self._sinusoid_model - cur_model_i + new_model

        # update the sinusoid parameters
        self._f_n[indices] = f_n_new
        self._a_n[indices] = a_n_new
        self._ph_n[indices] = ph_n_new
        self._include[indices] = True  # we assume that the changed sinusoids need to be included

        # update harmonic parameters
        self._harmonics[indices] = harmonics_new
        self._h_base[indices] = h_base_new
        self._h_mult[indices] = h_mult_new

        # update numbers
        self.update_n()

        return None

    def remove_sinusoids(self, time, indices):
        """Remove the sinusoids at the provided indices from the list.

        Meant for updating a limited number of sinusoids, less efficient for large numbers.
        For that case, see set_sinusoids.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        indices: numpy.ndarray[Any, dtype[int]]
            Indices of the sinusoids to remove.
        """
        indices = np.atleast_1d(indices)

        # get a list of indices that are currently included in the model
        i_include = indices[self._include[indices]]

        # get the current model at the indices
        cur_model_i = sum_sines_st(time, self._f_n[i_include], self._a_n[i_include], self._ph_n[i_include])

        # update the model
        self._sinusoid_model = self._sinusoid_model - cur_model_i

        # update the sinusoid parameters
        self._f_n = np.delete(self._f_n, indices)
        self._a_n = np.delete(self._a_n, indices)
        self._ph_n = np.delete(self._ph_n, indices)
        self._include = np.delete(self._include, indices)

        # update harmonic parameters
        self._harmonics = np.delete(self._harmonics, indices)
        self._h_base = np.delete(self._h_base, indices)
        self._h_mult = np.delete(self._h_mult, indices)

        # if we deleted a base harmonic, also delete the harmonic series
        self._check_removed_h_base(indices)

        # update indices for the removals (needs to go after _check_removed_h_base)
        self._h_base = np.array(ut.adjust_indices_removed(self._h_base, indices))

        # update numbers
        self.update_n()

        return None

    def include_sinusoids(self, time, indices):
        """Add back the sinusoids at the provided indices to the model.

        Meant for updating a limited number of sinusoids, less efficient for large numbers.
        For that case, see set_sinusoids.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        indices: numpy.ndarray[Any, dtype[int]]
            Indices of the sinusoids to include.
        """
        indices = np.atleast_1d(indices)

        # get a list of indices that are currently excluded from the model
        i_exclude = indices[~self._include[indices]]

        # get the current model at the indices
        cur_model_i = sum_sines_st(time, self._f_n[i_exclude], self._a_n[i_exclude], self._ph_n[i_exclude])

        # update the model
        self._sinusoid_model = self._sinusoid_model + cur_model_i

        # set their include parameter
        self._include[i_exclude] = True

        # update numbers
        self.update_n()

        return None

    def exclude_sinusoids(self, time, indices):
        """Remove the sinusoids at the provided indices from the model.

        Does not remove the sinusoids from the list yet, for index consistency.
        Meant for updating a limited number of sinusoids, less efficient for large numbers.
        For that case, see set_sinusoids.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series
        indices: numpy.ndarray[Any, dtype[int]]
            Indices of the sinusoids to exclude.
        """
        indices = np.atleast_1d(indices)

        # get a list of indices that are currently included in the model
        i_include = indices[self._include[indices]]

        # get the current model at the indices
        cur_model_i = sum_sines_st(time, self._f_n[i_include], self._a_n[i_include], self._ph_n[i_include])

        # update the model
        self._sinusoid_model = self._sinusoid_model - cur_model_i

        # set their include parameter
        self._include[i_include] = False

        # update numbers
        self.update_n()

        return None

    def remove_excluded(self):
        """Remove the sinusoids that are currently not included from the list."""
        # make removal indices out of _include
        indices = np.arange(len(self._include))[~self._include]

        # the model already doesn't include these sinusoids

        # remove the sinusoid parameters
        self._f_n = self._f_n[self._include]
        self._a_n = self._a_n[self._include]
        self._ph_n = self._ph_n[self._include]

        # remove the harmonic parameters
        self._harmonics = self._harmonics[self._include]
        self._h_base = self._h_base[self._include]
        self._h_mult = self._h_mult[self._include]

        # include itself needs to go last
        self._include = self._include[self._include]

        # check for base frequency removal
        self._check_removed_h_base(indices)

        # update indices for the removals (needs to go after _check_removed_h_base)
        self._h_base = np.array(ut.adjust_indices_removed(self._h_base, indices))

        # numbers do not change here, except if we removed a base frequency
        self.update_n()  # n_sin won't change but n_harm and n_base might

        return None

    def update_sinusoid_uncertainties(self, time, residual, flux_err, i_chunks):
        """Update the sinusoid model parameter errors using the residual flux.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        residual: numpy.ndarray[Any, dtype[float]]
            Residual flux (flux minus model).
        flux_err: numpy.ndarray[Any, dtype[float]]
            Errors in the measurement values.
        i_chunks: numpy.ndarray[Any, dtype[int]]
            Pair(s) of indices indicating time chunks within the light curve, separately handled in cases like
            the piecewise-linear curve. If only a single curve is wanted, set to np.array([[0, len(time)]]).
        """
        out = ut.formal_uncertainties_sinusoid(time, residual, flux_err, self._a_n, i_chunks)

        self._f_n_err = out[0]
        self._a_n_err = out[1]
        self._ph_n_err = out[2]

        return

    def update_sinusoid_uncertainties_harmonic(self):
        """Update the harmonic sinusoid model parameter errors.

        Note: uses the sinusoid uncertainties. These should be updated first.
        """
        self._f_h_err = ut.formal_uncertainties_harmonic(self._f_n_err, self._h_base, self._h_mult)

        return

    def update_sinusoid_passing_sigma(self):
        """Update the passing status of the sinusoids for the sigma criterion.

        Note: uses the sinusoid uncertainties. These should be updated first.
        """
        # find the insignificant frequencies
        out = frs.remove_insignificant_sigma(self._f_n, self._f_n_err, self._a_n, self._a_n_err, sigma_f=3, sigma_a=3)

        # invert removal mask
        self._passing_sigma = ~out

        return

    def update_sinusoid_passing_snr(self, time, residual, window_width=1.):
        """Update the passing status of the sinusoids for the sigma criterion.

        Parameters
        ----------
        time: numpy.ndarray[Any, dtype[float]]
            Timestamps of the time series.
        residual: numpy.ndarray[Any, dtype[float]]
            Residual flux (flux minus model).
        window_width: float
            The width of the window used to compute the noise spectrum.
        """
        # calculate the noise at each frequency
        noise_at_f = pdg.scargle_noise_at_freq(self._f_n, time, residual, window_width=window_width)

        # determine the insignificant frequencies
        out = frs.remove_insignificant_snr(time, self._a_n, noise_at_f)

        # invert removal mask
        self._passing_snr = ~out

        return

    def update_sinusoid_passing_harmonic(self, f_resolution):
        """Update the passing status of the sinusoids for the sigma criterion.

        Parameters
        ----------
        f_resolution: float
            Frequency resolution of the time series
        """
        # select candidate harmonic frequencies meeting some criteria
        self._passing_harmonic = np.zeros(self.n_sin, dtype=bool)
        for f_base in self.f_base:
            harmonics, _ = frs.select_harmonics_sigma(self._f_n, self._f_n_err, f_base, f_tol=f_resolution / 2,
                                                      sigma_f=3)
            self._passing_harmonic[harmonics] = True

        return



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the main functions that link together all functionality.

Code written by: Luc IJspeert
"""
from star_shine.config.helpers import get_config, get_config_path


# load configuration
config = get_config()
# todo: modelling red-noise with autoregressive model
# todo: amplitude factor for each sector (linear model)
# todo: consider twice nyquist for harmonics
# todo: enforce_harmonics do linear fit


def update_config(file_name='', settings=None):
    """Update the configuration using a file and/or a dictionary.

    First loads the file, then updates settings, so both could be used simultaneously.
    This alters the state of the current configuration, not the configuration file.

    Parameters
    ----------
    file_name: str, optional
        Path to the yaml configuration file.
    settings: dict, optional
        Dictionary to update specific configuration settings.

    Returns
    -------
    None
    """
    # load from file
    if file_name != '':
        config.update_from_file(file_name)

    # update individual settings
    if settings is not None:
        config.update_from_dict(settings)

    return None

def save_config(file_name=''):
    """Save the configuration to a file."""
    # if no file name is supplied, overwrite it in the default place
    if file_name == '':
        file_name = get_config_path()

    config.save_to_file(file_name)

    return None



"""STAR SHINE
Satellite Time-series Analysis Routine using Sinusoids and Harmonics through Iterative Non-linear Extraction

This Python module contains the data class for handling the user defined data to analyse.

Code written by: Luc IJspeert
"""
import os
import numpy as np

from star_shine.core import time_series as tms, utility as ut, visualisation as vis
from star_shine.core import io
from star_shine.config.helpers import get_config


# load configuration
config = get_config()


class Data:
    """A class to handle light curve data.

    Attributes
    ----------
    file_list: list[str]
        List of ascii light curve files or (TESS) data product '.fits' files.
    data_dir: str
        Root directory where the data files to be analysed are located.
    target_id: str
        User defined identification integer for the target under investigation.
    data_id: str
        User defined identification for the dataset used.
    flux_counts_medians: numpy.ndarray[Any, dtype[float]]
        Median flux counts per chunk.
    """

    def __init__(self, target_id='', data_id=''):
        """Initialises the Data object.

        The data is loaded from the given file(s) and some basic processing is done.
        Either a file name, or target id plus file list must be given.

        Parameters
        ----------
        target_id: str, optional
            User defined identification number or name for the target under investigation. If empty, the file name
            of the first file in file_list is used.
        data_id: str, optional
            User defined identification name for the dataset used.
        """
        self.file_list = []
        self.data_dir = ''
        self.target_id = target_id
        self.data_id = data_id

        # the time series is stored here
        self.time_series = None

        # additional time series properties not in time_series
        self.flux_counts_medians = np.zeros((0,))

        return

    def __repr__(self):
        return (f"Data(target_id={self.target_id!r}, "
                f"data_id={self.data_id!r}, "
                f"data_dir={self.data_dir!r}, "
                f"file_list={self.file_list!r})")

    def _check_file_existence(self, logger=None):
        """Checks whether the given file(s) exist.

        Removes missing files from the file list

        Parameters
        ----------
        logger: logging.Logger, optional
            Instance of the logging library.
        """
        # check for missing files in the list
        missing = []
        for i, file in enumerate(self.file_list):
            if not os.path.exists(os.path.join(self.data_dir, file)):
                missing.append(i)

        # log a message if files are missing
        if len(missing) > 0:
            missing_files = [self.file_list[i] for i in missing]

            # add directory to message
            dir_text = ""
            if self.data_dir is not None:
                dir_text = f" in directory {self.data_dir}"
            message = f"Missing files {missing_files}{dir_text}, removing from list."

            if logger is not None:
                logger.warning(message)

            # remove the files
            for i in missing:
                del self.file_list[i]

        return None

    def get_dict(self):
        """Make a dictionary of the attributes.

        Primarily for saving to file.

        Returns
        -------
        data_dict: dict
            Dictionary of the data attributes and fields
        """
        # make a dictionary of the fields to be saved
        data_dict = dict()
        data_dict['target_id'] = self.target_id
        data_dict['data_id'] = self.data_id
        data_dict['date_time'] = ut.datetime_formatted()

        # original list of files
        data_dict['data_dir'] = self.data_dir
        data_dict['file_list'] = self.file_list

        # summary statistics
        data_dict['t_tot'] = self.time_series.t_tot
        data_dict['t_mean'] = self.time_series.t_mean
        data_dict['t_mean_chunk'] = self.time_series.t_mean_chunk
        data_dict['t_step'] = self.time_series.t_step

        # the time series data
        data_dict['time'] = self.time_series.time
        data_dict['flux'] = self.time_series.flux
        data_dict['flux_err'] = self.time_series.flux_err

        # additional information
        data_dict['i_chunks'] = self.time_series.i_chunks
        data_dict['flux_counts_medians'] = self.flux_counts_medians

        return data_dict

    @classmethod
    def load_data(cls, file_list, data_dir='', target_id='', data_id='', logger=None):
        """Load light curve data from the file list.

        Parameters
        ----------
        file_list: list[str]
            List of ascii light curve files or (TESS) data product '.fits' files. Exclude the path given to 'data_dir'.
            If only one file is given, its file name is used for target_id (if 'none').
        data_dir: str, optional
            Root directory where the data files to be analysed are located. Added to the file name.
            If empty, it is loaded from config.
        target_id: str, optional
            User defined identification number or name for the target under investigation. If empty, the file name
            of the first file in file_list is used.
        data_id: str, optional
            User defined identification name for the dataset used.
        logger: logging.Logger, optional
            Instance of the logging library.

        Returns
        -------
        Data
            Instance of the Data class with the loaded data.
        """
        instance = cls()

        # set the file list and data directory
        if data_dir == '':
            data_dir = config.data_dir
        instance.file_list = file_list
        instance.data_dir = data_dir

        # guard against empty list
        if len(file_list) == 0:
            if logger is not None:
                logger.warning("Empty file list provided.")
            return cls()

        # Check if the file(s) exist(s)
        instance._check_file_existence(logger=logger)
        if len(instance.file_list) == 0:
            if logger is not None:
                logger.warning("No existing files in file list")
            return cls()

        # set IDs
        if target_id == '':
            target_id = os.path.splitext(os.path.basename(file_list[0]))[0]  # file name is used as identifier
        instance.target_id = target_id
        instance.data_id = data_id

        # add data_dir for loading files, if not None
        if instance.data_dir == '':
            file_list_dir = instance.file_list
        else:
            file_list_dir = [os.path.join(instance.data_dir, file) for file in instance.file_list]

        # load the data from the list of files
        lc_data = io.load_light_curve(file_list_dir, apply_flags=config.apply_q_flags)

        # make a TimeSeries instance
        instance.time_series = tms.TimeSeries(lc_data[0], lc_data[1], lc_data[2], lc_data[3])
        instance.flux_counts_medians = lc_data[4]

        # check for overlapping time stamps
        if np.any(np.diff(instance.time_series.time) <= 0) & (logger is not None):
            logger.warning("The time array chunks include overlap.")

        if logger is not None:
            logger.info(f"Loaded data from external file(s).")

        return instance

    @classmethod
    def load(cls, file_name, data_dir='', h5py_file_kwargs=None, logger=None):
        """Load a data file in hdf5 format.

        Parameters
        ----------
        file_name: str
            File name to load the data from
        data_dir: str, optional
            Root directory where the data files to be analysed are located. Added to the file name.
            If empty, it is loaded from config.
        h5py_file_kwargs: dict, optional
            Keyword arguments for opening the h5py file.
            Example: {'locking': False}, for a drive that does not support locking.
        logger: logging.Logger, optional
            Instance of the logging library.

        Returns
        -------
        Data
            Instance of the Data class with the loaded data.
        """
        # initiate the Data instance
        instance = cls()

        # set the file list and data directory
        if data_dir == '':
            data_dir = config.data_dir
        instance.data_dir = data_dir

        # add data_dir for loading file, if not None
        if instance.data_dir != '':
            file_name = os.path.join(instance.data_dir, file_name)

        # io module handles opening the file
        data_dict = io.load_data_hdf5(file_name, h5py_file_kwargs=h5py_file_kwargs)

        # general properties
        instance.target_id = data_dict['target_id']
        instance.data_id = data_dict['data_id']
        instance.file_list = data_dict['file_list']

        # make a TimeSeries instance
        instance.time_series = tms.TimeSeries(data_dict['time'], data_dict['flux'], data_dict['flux'],
                                              data_dict['i_chunks'])
        instance.flux_counts_medians = data_dict['flux_counts_medians']

        if logger is not None:
            logger.info(f"Loaded data file with target identifier: {data_dict['target_id']}, "
                        f"created on {data_dict['date_time']}. Data identifier: {data_dict['data_id']}.")

        return instance

    def save(self, file_name):
        """Save the data to a file in hdf5 format.

        Parameters
        ----------
        file_name: str
            File name to save the data to
        """
        # make a dictionary of the fields to be saved
        data_dict = self.get_dict()

        # io module handles writing to file
        io.save_data_hdf5(file_name, data_dict)

        return None

    def plot_time_series(self, file_name=None, show=True):
        """Plot the time series data.

        Parameters
        ----------
        file_name: str, optional
            File path to save the plot
        show: bool, optional
            If True, display the plot
        """
        vis.plot_lc(self.time_series.time, self.time_series.flux, self.time_series.flux_err, self.time_series.i_chunks,
                    file_name=file_name, show=show)

        return None

    def plot_periodogram(self, plot_per_chunk=False, file_name=None, show=True):
        """Plot the periodogram of the time series.

        Parameters
        ----------
        plot_per_chunk: bool
            If True, plot the periodogram of each separate time chunk in one plot.
        file_name: str, optional
            File path to save the plot
        show: bool, optional
            If True, display the plot
        """
        vis.plot_pd(self.time_series.time, self.time_series.flux, self.time_series.i_chunks,
                    plot_per_chunk=plot_per_chunk, file_name=file_name, show=show)

        return None



"""STAR SHINE __init__ file

Code written by: Luc IJspeert
"""

from .api.main import *
from .api.data import Data
from .api.result import Result
from .api.pipeline import Pipeline

try:
    # GUI
    from .gui.gui_app import launch_gui
except ImportError:
    print('GUI unavailable, likely missing dependency PySide6.')
    pass

__all__ = ['gui', 'api', 'core', 'config', 'data']
